# -*- coding: utf-8 -*-
import logging
import sys

from nbsafety.singletons import nbs
from pyccolo.extra_builtins import EMIT_EVENT
from .utils import assert_bool, make_safety_fixture, skipif_known_failing

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture()


def run_cell(cell, **kwargs):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell, **kwargs)


def stale_detected():
    return nbs().test_and_clear_detected_flag()


def assert_detected(msg=""):
    assert_bool(stale_detected(), msg=msg)


def assert_false_positive(msg=""):
    """
    Same as `assert_detected` but asserts a false positive.
    Helps with searchability of false positives in case we want to fix these later.
    """
    return assert_detected(msg=msg)


def assert_not_detected(msg=""):
    assert_bool(not stale_detected(), msg=msg)


def assert_false_negative(msg=""):
    """
    Same as `assert_not_detected` but asserts a false negative.
    Helps with searchability of false negatives in case we want to fix these later.
    """
    return assert_not_detected(msg=msg)


def test_simplest():
    run_cell("a = 1")
    run_cell("b = a + 1")
    run_cell("a = 3")
    run_cell("logging.info(b)")
    assert_detected("should have detected b has stale dep on old a")


def test_readme_example():
    run_cell("def eval_model_1(): return 0.5")
    run_cell("def eval_model_2(): return 0.85")
    run_cell('models = {"model_1": eval_model_1, "model_2": eval_model_2}')
    output = """
        best_acc, best_model = max((f(), name) for name, f in models.items())
        logging.info(f'The best model was {best_model} with an accuracy of {best_acc}.')
        """
    run_cell(output)
    run_cell("def eval_model_1(): return 0.9")
    run_cell(output)
    assert_detected("`models` depends on stale `eval_model_1`")


# TODO: to get this working properly, post-call argument
#  symbols need to have dependencies on pre-call argument
#  symbols. The place to do this is in `DataSymbol.create_symbols_for_call_args(...)`
def test_passed_sym_captured_as_dep_for_mutated_obj():
    run_cell(
        """
        class Foo:
            def __init__(self, x):
                self.x = x
                
        def mutate(foo, x):
            foo.x = x
        """
    )
    run_cell("foo = Foo(5)")
    run_cell("y = 7")
    run_cell("mutate(foo, y)")
    run_cell("logging.info(foo.x)")
    assert_not_detected()
    run_cell("y = 42")
    run_cell("logging.info(foo.x)")
    assert_detected("`foo.x` depends on old value of `y`")


def test_subscript_dependency():
    run_cell("lst = [0, 1, 2]")
    run_cell("x = 5")
    run_cell("y = x + lst[0]")
    run_cell("lst[0] = 10")
    run_cell("logging.info(y)")
    assert_detected("`y` depends on stale `lst[0]`")


def test_long_chain():
    run_cell("a = 1")
    run_cell("b = a + 1")
    run_cell("c = b + 1")
    run_cell("d = c + 1")
    run_cell("e = d + 1")
    run_cell("f = e + 1")
    assert_not_detected("everything OK so far")
    run_cell("a = 2")
    run_cell("logging.info(f)")
    assert_detected("f has stale dependency on old value of a")


def test_for_loop_liveness_check():
    run_cell("x = 0")
    run_cell("y = x + 1")
    run_cell("x = 42")
    run_cell(
        """
for i in range(2):
    logging.info(y)
    y = 7
"""
    )
    assert_detected("first use of `y` in loop has stale dependency on old value of `x`")


def test_redef_after_stale_use():
    run_cell("a = 1")
    run_cell("b = a + 1")
    run_cell("a = 3")
    run_cell(
        """
        logging.info(b)
        b = 7
        """
    )
    assert_detected("b has stale dependency on old value of a")


def test_class_redef():
    classdef = """
        class Foo:
            def __init__(self, x):
                self.x = x
        """
    run_cell(classdef)
    run_cell("x = 5")
    run_cell("foo = Foo(x)")
    run_cell(classdef)
    run_cell("logging.info(x)")
    assert_not_detected("`x` independent of class `Foo`")


def test_subscript_dependency_fp():
    run_cell("lst = [0, 1, 2]")
    run_cell("x = 5")
    run_cell("y = x + lst[0]")
    run_cell("lst[1] = 10")
    run_cell("logging.info(y)")
    assert_not_detected("y depends only on unchanged lst[0] and not on changed lst[1]")


def test_comprehension_generator_vars_not_live():
    run_cell("x = 0")
    run_cell("y = x + 6")
    run_cell("x = 42")
    run_cell("lst = [y for y in range(1) for j in range(1)]")
    assert_not_detected("`y` is not live in the list comprehension")


def test_lambda_params_not_live():
    run_cell("x = 0")
    run_cell("y = x + 6")
    run_cell("x = 42")
    run_cell("f = lambda y: y * 2")
    assert_not_detected("`y` is not live in the lambda expr")


def test_lambda_arg():
    run_cell("x = 5")
    run_cell("y = 7")
    run_cell("lam = lambda x: x * 2")
    run_cell("z = lam(y + 3)")
    run_cell("x = 42")
    run_cell("logging.info(z)")
    assert_not_detected("`z` independent of updated `x`")
    run_cell("y = 43")
    run_cell("logging.info(z)")
    assert_detected("`z` depends on old value of `y`")


def test_lambda_scope():
    run_cell(
        """
        def foo(x):
            return lambda: x + 5
        """
    )
    run_cell("x = 5")
    run_cell("y = 7")
    run_cell("lam = foo(y)")
    run_cell("z = lam()")
    run_cell("x = 42")
    run_cell("logging.info(z)")
    assert_not_detected("`z` independent of updated `x`")
    run_cell("y = 43")
    run_cell("logging.info(z)")


def test_lambda_wrapping_call():
    run_cell("z = 42")
    run_cell(
        """
        def f():
            return z
        """
    )
    run_cell("lam = lambda: f()")
    run_cell("x = lam()")
    run_cell("z = 43")
    run_cell("logging.info(x)")
    assert_detected("`x` depends on old value of `z`")


def test_lambda_with_kwarg_scope():
    run_cell("w = 99")
    run_cell(
        """
        def foo(x):
            lam = lambda t=w: t + x + 5
            return lam
        """
    )
    run_cell("x = 5")
    run_cell("y = 7")
    run_cell("lam = foo(y)")
    run_cell("z = lam()")
    run_cell("x = 42")
    run_cell("logging.info(z)")
    assert_not_detected("`z` independent of updated `x`")
    run_cell("w = 43")
    run_cell("logging.info(z)")
    assert_detected("`z` depends on old value of `y`")


def test_fundef_params_not_live():
    run_cell("x = 0")
    run_cell("y = x + 6")
    run_cell("x = 42")
    run_cell(
        """
        def f(y):
            return y + 9
        """
    )
    assert_not_detected("`y` is not live in the function")


def test_no_edge_for_comprehension_variable():
    run_cell("i = 0")
    run_cell("x = [i for i in range(10)]")
    run_cell("i = 1")
    run_cell("logging.info(x)")
    assert_not_detected("`x` does not depend on `i`")


# simple test about the basic assignment
def test_basic_assignment():
    run_cell("a = 1")
    run_cell("b = 2")
    run_cell("c = a+b")
    run_cell("d = c+1")
    run_cell('logging.info("%s, %s, %s, %s", a, b, c, d)')
    # redefine a here but not c and d
    run_cell("a = 7")
    run_cell('logging.info("%s, %s, %s, %s", a, b, c, d)')
    assert_detected("Did not detect that c's reference was changed")

    run_cell("c = a+b")
    run_cell('logging.info("%s, %s, %s, %s", a, b, c, d)')
    assert_detected("Did not detect that d's reference was changed")

    run_cell("d = c+1")
    run_cell('logging.info("%s, %s, %s, %s", a, b, c, d)')
    assert_not_detected("There should be no more dependency issue")


def test_empty_list_assignment():
    run_cell("a = [5]")
    run_cell("b = a + [6]")
    run_cell("logging.info(b)")
    run_cell("a = [6]")
    run_cell("logging.info(b)")
    assert_detected("`b` depends on stale `a`")
    run_cell("b = a + [7]")
    run_cell("a = []")
    run_cell("logging.info(b)")
    assert_detected("`b` depends on stale `a`")


# redefined function example from the project prompt
def test_redefined_function_in_list():
    run_cell(
        """
        def foo():
            return 5

        def bar():
            return 7
        """
    )
    run_cell(
        """
        funcs_to_run = [foo, bar]
        """
    )
    run_cell(
        """
        accum = 0
        for f in funcs_to_run:
            accum += f()
        logging.info(accum)
        """
    )

    # redefine foo here but not funcs_to_run
    run_cell(
        """
        def foo():
            return 10

        def bar():
            return 7
        """
    )
    run_cell(
        """
        accum = 0
        for f in funcs_to_run:
            accum += f()
        logging.info(accum)
        """
    )
    assert_detected("Did not detect that funcs_to_run's reference was changed")

    run_cell(
        """
        funcs_to_run = [foo,bar]
        """
    )
    run_cell(
        """
        accum = 0
        for f in funcs_to_run:
            accum += f()
        logging.info(accum)
        """
    )
    assert_not_detected("There should be no more dependency issue")


# like before but the function is called in a list comprehension
def test_redefined_function_for_funcall_in_list_comp():
    run_cell(
        """
        def foo():
            return 5

        def bar():
            return 7
        """
    )
    run_cell("retvals = [foo(), bar()]")
    run_cell(
        """
        accum = 0
        for ret in retvals:
            accum += ret
        logging.info(accum)
        """
    )

    # redefine foo here but not funcs_to_run
    run_cell(
        """
        def foo():
            return 10

        def bar():
            return 7
        """
    )
    run_cell("logging.info(accum)")
    assert_detected("Did not detect stale dependency of `accum` on `foo` and `bar`")


# like before but we run the list through a function before iterating
def test_redefined_function_for_funcall_in_modified_list_comp():
    run_cell(
        """
        def foo():
            return 5

        def bar():
            return 7
        """
    )
    run_cell("retvals = tuple([foo(), bar()])")
    run_cell(
        """
        accum = 0
        # for ret in map(lambda x: x * 5, retvals):
        for ret in retvals:
            accum += ret
        logging.info(accum)
        """
    )

    # redefine foo here but not funcs_to_run
    run_cell(
        """
        def foo():
            return 10

        def bar():
            return 7
        """
    )
    run_cell("logging.info(accum)")
    assert_detected("Did not detect stale dependency of `accum` on `foo` and `bar`")


def test_for_loop_with_map():
    run_cell(
        """
        accum = 0
        foo = [1, 2, 3, 4, 5]
        for ret in map(lambda x: x * 5, foo):
            accum += ret
        """
    )
    run_cell("logging.info(accum)")
    assert_not_detected("no stale dep foo -> accum")
    run_cell("foo = [0]")
    run_cell("logging.info(accum)")
    assert_detected(
        "should detect stale dep foo -> accum unless only propagating past cell bounds"
    )


def test_redefined_function_over_list_comp():
    run_cell(
        """
        def foo():
            return 5

        def bar():
            return 7

        def baz(lst):
            return map(lambda x: 3*x, lst)
        """
    )
    run_cell("retvals = baz([foo(), bar()])")
    run_cell(
        """
        accum = 0
        for ret in map(lambda x: x * 5, retvals):
            accum += ret
        """
    )
    run_cell(
        """
        def baz(lst):
            return map(lambda x: 7*x, lst)
        """
    )
    run_cell("logging.info(accum)")
    assert_detected("Did not detect stale dependency of `accum` on `baz`")


# like before but the function is called in a tuple comprehension
def test_redefined_function_for_funcall_in_tuple_comp():
    run_cell(
        """
        def foo():
            return 5

        def bar():
            return 7
        """
    )
    run_cell("retvals = (foo(), bar())")
    run_cell(
        """
        accum = 0
        for ret in retvals:
            accum += ret
        logging.info(accum)
        """
    )

    # redefine foo here but not funcs_to_run
    run_cell(
        """
        def foo():
            return 10

        def bar():
            return 7
        """
    )
    run_cell("logging.info(accum)")
    assert_detected("Did not detect stale dependency of `accum` on `foo` and `bar`")


def test_symbol_callpoint():
    run_cell("x = 0")
    run_cell("y = x + 1")
    run_cell(
        """
def f():
    return y + 7
"""
    )
    run_cell("z = f()")
    assert_not_detected()
    run_cell("x = 42")
    run_cell("z = f()")
    assert_detected("`y` as referenced in call of function `f` is stale")


def test_symbol_callpoint_2():
    run_cell("x = 0")
    run_cell("y = x + 1")
    run_cell(
        """
        def f():
            y = 9
            return y + 7
        """
    )
    run_cell("z = f()")
    assert_not_detected()
    run_cell("x = 42")
    run_cell("z = f()")
    assert_not_detected()


def test_symbol_callpoint_3():
    run_cell("x = 0")
    run_cell("y = x + 1")
    run_cell(
        """
        def f():
            return y + 7
        """
    )
    run_cell("z = f()")
    assert_not_detected()
    run_cell("x = 42")
    run_cell(
        """
        def f():
            return y + 7
        """
    )
    run_cell("z = f()")
    assert_detected("`y` as referenced in call of function `f` is stale")


def test_symbol_callpoint_4():
    run_cell("x = 0")
    run_cell("y = x + 1")
    run_cell(
        """
        def f(y):
            return y + 7
        """
    )
    run_cell("z = f(5)")
    assert_not_detected()
    run_cell("x = 42")
    run_cell("z = f(5)")
    assert_not_detected(
        "`y` as referenced in call of function `f` is an arg and therefore not stale"
    )


def test_function_arg_independent_of_outer_staleness():
    run_cell("x = 0")
    run_cell("y = x + 1")
    run_cell(
        """
        def f(y):
            return y + 7
        """
    )
    run_cell("z = f(5)")
    assert_not_detected()
    run_cell("x = 42")
    run_cell("logging.info(z)")
    assert_not_detected("`z` independent of `x`")


def test_lambda_capture():
    run_cell("x = 42")
    run_cell("lst = []")
    run_cell("lst.append(lambda elt: elt + x)")
    run_cell("y = lst[0](7)")
    run_cell("x = 43")
    run_cell("logging.info(y)")
    assert_detected("`y` depends on stale `x`")


# Tests about variables that have same name but in different scope.
# There shouldn't be any extra dependency because of the name.
def test_variable_scope():
    run_cell(
        """
        def func():
            x = 6
        """
    )
    run_cell("x = 7")
    run_cell("y = x")
    run_cell("z = func")
    run_cell('logging.info("%s, %s", y, z())')

    # change x inside of the function, but not x outside of the function
    run_cell("def func():\n    x = 10")
    run_cell('logging.info("%s, %s", y, z())')
    assert_detected("Did not detect the dependency change in the function")

    run_cell("y = x")
    run_cell('logging.info("%s, %s", y, z())')
    assert_detected(
        "Updating y should not solve the dependency change inside of function func"
    )

    run_cell("z = func")
    run_cell('logging.info("%s, %s", y, z())')
    assert_not_detected("Updating z should solve the problem")


def test_variable_scope_2():
    run_cell("def func():\n    x = 6")
    run_cell("x = 7")
    run_cell("y = x")
    run_cell("z = func")
    run_cell('logging.info("%s, %s", y, z())')

    # change x outside of the function, but not inside of the function
    run_cell("x = 10")
    run_cell('logging.info("%s, %s", y, z())')
    assert_detected("Did not detect the dependency change outside of the function")

    run_cell("z = func")
    run_cell('logging.info("%s, %s", y, z())')
    assert_detected(
        "Updating z should not solve the dependency change outside of function"
    )

    run_cell("y = x")
    run_cell('logging.info("%s, %s", y, z())')
    assert_not_detected("Updating y should solve the problem")


def test_default_args():
    run_cell(
        """
        x = 7
        def foo(y=x):
            return y + 5
        """
    )
    run_cell("a = foo()")
    assert_not_detected()
    run_cell("x = 10")
    assert_not_detected()
    run_cell("b = foo()")
    assert_detected("Should have detected stale dependency of fn foo() on x")


def test_same_pointer():
    # a and b are actually pointing to the same thing
    run_cell("a = [7]")
    run_cell("b = a")
    run_cell("c = b + [5]")

    run_cell("a[0] = 8")
    run_cell("logging.info(b)")
    assert_not_detected(
        "`b` is an alias of `a`, updating a should automatically update `b` as well"
    )
    run_cell("logging.info(c)")
    assert_detected(
        "`c` does not point to the same thing as `a` or `b`, thus there is a stale dependency here"
    )


def test_func_assign_objs():
    run_cell(
        """
        a = [1]
        b = [1]
        c = [2]
        d = [3]"""
    )
    run_cell(
        """
        def func(x, y=a):
            e = [c[0] + d[0]]
            f = [x[0] + y[0]]
            return f
        """
    )
    run_cell("z = func(c)")
    run_cell("a = [4]")
    run_cell("logging.info(z[0])")
    assert_detected("Should have detected stale dependency of fn func on a")
    run_cell(
        """
        def func(x, y=a):
            logging.info(b[0])
            e = [c[0] + d[0]]
            f = [x[0] + y[0]]
            %safety show_deps f[0]
            return f
        # z = func(c)
        """
    )
    run_cell("z = func(c)")
    run_cell("%safety show_deps z[0]")
    run_cell("logging.info(z[0])")
    assert_not_detected()
    run_cell("logging.info(z)")
    assert_not_detected()
    run_cell("c = [3]")
    run_cell("logging.info(z)")
    assert_detected("Should have detected stale dependency of z on c")
    run_cell("c = [77]")
    run_cell("logging.info(z[0])")
    assert_detected("Should have detected stale dependency of z on c")
    run_cell("z = func(c)")
    run_cell("logging.info(z[0])")
    assert_not_detected()
    run_cell("logging.info(z)")
    assert_not_detected()
    run_cell("b = [40]")
    run_cell("d = [4]")
    run_cell("logging.info(z[0])")
    assert_not_detected("Changing b and d should not affect z")
    run_cell("logging.info(z)")
    assert_not_detected("Changing b and d should not affect z")


def test_func_assign_ints():
    run_cell(
        """
        a = 1
        b = 1
        c = 2
        d = 3
        def func(x, y=a):
            e = c + d
            f = x + y
            return f
        """
    )
    run_cell("z = func(c)")
    run_cell("a = 4")
    run_cell("logging.info(z)")
    assert_detected("Should have detected stale dependency of fn func on a")
    run_cell(
        """
        def func(x, y=a):
            logging.info(b)
            e = c + d
            f = x + y
            return f
        z = func(c)
        """
    )
    run_cell("logging.info(z)")
    assert_not_detected()
    run_cell("c = 3")
    run_cell("logging.info(z)")
    assert_detected("Should have detected stale dependency of z on c")
    run_cell("z = func(c)")
    run_cell("logging.info(z)")
    assert_not_detected()
    run_cell("b = 4")
    run_cell("d = 1")
    assert_not_detected("Changing b and d should not affect z")


def test_func_assign_helper_func():
    run_cell(
        """
        x = 3
        a = 4
        def f():
            def g():
                logging.info(a)
                return x
            return g()
        y = f()
        """
    )
    run_cell("x = 4")
    run_cell("logging.info(y)")
    assert_detected("Should have detected stale dependency of y on x")
    run_cell("y = f()")
    run_cell("logging.info(y)")
    assert_not_detected()
    run_cell("a = 1")
    run_cell("logging.info(y)")
    assert_not_detected("Changing a should not affect y")


def test_func_assign_helper_func_2():
    run_cell(
        """
        x = 3
        a = 4
        def f():
            def g():
                logging.info(a)
                return x
            return g
        y = f()()
        """
    )
    run_cell("x = 4")
    run_cell("logging.info(y)")
    assert_detected("Should have detected stale dependency of y on x")


def test_branching():
    run_cell("y = 7")
    run_cell("x = y + 3")
    run_cell(
        """
        if True:
            b = 5
        else:
            y = 7
        """
    )
    run_cell("logging.info(x)")
    assert_not_detected("false positive on unchanged y")


def test_branching_2():
    run_cell("y = 7")
    run_cell("x = y + 3")
    run_cell(
        """
        if False:
            b = 5
        else:
            y = 9
        """
    )
    run_cell("logging.info(x)")
    assert_detected("x depends on stale y")


def test_identity_checking():
    run_cell("y = 7")
    run_cell("x = y + 3")
    run_cell("y = 7")
    run_cell("logging.info(x)")
    assert_not_detected("`y` was not mutated")


def test_identity_checking_obj():
    # To get this working properly, we need to create datasyms for all of the values in the literal
    run_cell("y = [7]")
    run_cell("x = y + [3]")
    run_cell("y[0] = 7")
    run_cell("logging.info(x)")
    assert_not_detected("`y` was not mutated")


def test_identity_checking_obj_2():
    run_cell("x, y = [7], [8]")
    run_cell("z = y + [3]")
    run_cell("x[0] = 99")
    run_cell("logging.info(z)")
    assert_not_detected("`z` independent of x")
    run_cell("y[0] = 8")
    run_cell("logging.info(z)")
    assert_not_detected("`y` was not mutated")
    run_cell("y[0] = 42")
    run_cell("logging.info(z)")
    assert_detected("`y` was mutated")


def test_identity_checking_obj_3():
    run_cell('d = {"y": 7}')
    run_cell("x = list(d.values()) + [3]")
    run_cell('d["y"] = 7')
    run_cell("logging.info(x)")
    assert_not_detected('`d["y"]` was not mutated')
    run_cell('d["y"] = 8')
    run_cell("logging.info(x)")
    assert_detected('`d["y"]` was mutated')


def test_identity_checking_obj_4():
    # To get this working properly, we need to create datasyms for literal namespaces recursively
    run_cell('d = {"y": [7]}')
    run_cell('d["x"] = d["y"] + [3]')
    run_cell('d["y"][0] = 7')
    run_cell('logging.info(d["x"])')
    assert_not_detected('`d["y"]` was not mutated')
    run_cell('d["y"][0] = 8')
    run_cell('logging.info(d["x"])')
    assert_detected('`d["y"]` was mutated')


def test_identity_checking_obj_5():
    # To get this working properly, we need to create datasyms for literal namespaces recursively
    run_cell('d = {"y": {0: 7}}')
    run_cell('d["x"] = {1:3, **d["y"]}')
    run_cell('d["y"][0] = 7')
    run_cell('logging.info(d["x"])')
    assert_not_detected('`d["y"]` was not mutated')
    run_cell('d["y"][0] = 8')
    run_cell('logging.info(d["x"])')
    assert_detected('`d["y"]` was mutated')


def test_identity_checking_obj_6():
    # To get this working properly, we need to create datasyms for literal namespaces recursively
    run_cell("lst = [[1, 2], 0]")
    run_cell("lst[1] = lst[0] + [3, 4]")
    run_cell("lst[0][1] = 2")
    run_cell("logging.info(lst[1])")
    assert_not_detected("`lst[0][1]` was not mutated")
    run_cell("lst[0][1] = 42")
    run_cell("logging.info(lst[1])")
    assert_detected("`lst[0][1]` was mutated")


def test_starred_assignment_rhs():
    run_cell("x = 0")
    run_cell("y = 1")
    run_cell("z = 2")
    run_cell('lst = ["foo", "bar"]')
    # just to make sure the tracer can handle a starred expr in list literal
    run_cell("a, b, c, d, e = [x + 1, y + 2, z + 3, *lst]")
    run_cell("z = 42")
    run_cell("logging.info(a)")
    assert_not_detected()
    run_cell("logging.info(b)")
    assert_not_detected()
    run_cell("logging.info(c)")
    assert_detected()
    run_cell("logging.info(d)")
    assert_not_detected()
    run_cell("logging.info(e)")
    assert_not_detected()
    run_cell("lst[0] = 7")
    run_cell("logging.info(d)")
    assert_detected()
    run_cell("logging.info(e)")
    assert_not_detected()
    run_cell("lst[1] = 8")
    run_cell("logging.info(e)")
    assert_detected()


def test_starred_assignment():
    run_cell("x = 0")
    run_cell("y = 1")
    run_cell("z = 2")
    run_cell('lst = ["foo", "bar"]')
    # just to make sure the tracer can handle a starred expr in list literal
    run_cell("s, *t = [x + 1, y + 2, z + 3, *lst]")
    run_cell("z = 42")
    run_cell("logging.info(s)")
    assert_not_detected()
    run_cell("logging.info(t[0])")
    assert_not_detected()
    run_cell("logging.info(t[1])")
    assert_detected()
    run_cell("x = 99")
    run_cell("logging.info(s)")
    assert_detected()
    run_cell("logging.info(t[0])")
    assert_not_detected()
    run_cell("y = 142")
    run_cell("logging.info(t[0])")
    assert_detected()


def test_starred_assignment_in_middle():
    run_cell("a, b, c, d, e = 1, 2, 3, 4, 5")
    run_cell("x, *star, y = [a, b, c, d, e]")
    run_cell("a += 1")
    run_cell("logging.info(x)")
    assert_detected()
    run_cell("logging.info(star)")
    assert_not_detected()
    run_cell("logging.info(y)")
    assert_not_detected()
    run_cell("e += 1")
    run_cell("logging.info(y)")
    assert_detected()
    run_cell("logging.info(star)")
    assert_not_detected()
    run_cell("c += 1")
    run_cell("logging.info(star)")
    assert_detected()


def test_attributes():
    run_cell(
        """
        class Foo:
            def __init__(self, x):
                self.x = x
        """
    )
    run_cell("x = Foo(5)")
    run_cell("y = x.x + 5")
    run_cell("x.x = 8")
    run_cell("logging.info(y)")
    assert_detected("y depends on stale attrval x.x")


@skipif_known_failing
def test_attributes_2():
    run_cell(
        """
        class Foo:
            def __init__(self, x):
                self.x = x
        """
    )
    run_cell("x = Foo(5)")
    run_cell("y = x.x + 5")
    run_cell("%safety show_deps x")
    run_cell("%safety show_deps y")
    run_cell("x = 8")
    run_cell("logging.info(y)")
    assert_detected("y depends on stale x")


def test_attribute_unpacking():
    run_cell(
        """
        class Foo:
            def __init__(self, x):
                self.x = x
        """
    )
    run_cell("x = Foo(5)")
    run_cell("y = Foo(6)")
    run_cell("w = 42")
    run_cell("z = 43")
    run_cell("x.x, y.x = w + 2, z + 3")
    run_cell("z = 9001")
    run_cell("logging.info(x.x)")
    assert_not_detected()
    run_cell("logging.info(x)")
    assert_not_detected()
    run_cell("logging.info(y.x)")
    assert_detected()
    run_cell("logging.info(y)")
    assert_detected()
    run_cell("y.x = z + 3")
    run_cell("logging.info(y.x)")
    assert_not_detected()
    run_cell("w = 99")
    run_cell("logging.info(x.x)")
    assert_detected()
    run_cell("logging.info(x)")
    assert_detected()
    run_cell("logging.info(y.x)")
    assert_not_detected()
    run_cell("logging.info(y)")
    assert_not_detected()


def test_attribute_unpacking_no_overwrite():
    run_cell(
        """
        class Foo:
            def __init__(self, x):
                self.x = x
        """
    )
    run_cell("x = Foo(5)")
    run_cell("y = Foo(6)")
    run_cell("w = 42")
    run_cell("z = 43")
    run_cell("x.x, y.x = w + 2, z + 3")
    run_cell("s, t = 12, 13")
    run_cell("x.x, y.x = x.x + s, y.x + t")
    run_cell("w = 101")
    run_cell("logging.info(x.x)")
    assert_detected()
    run_cell("logging.info(y.x)")
    assert_not_detected()
    run_cell("z = 103")
    run_cell("logging.info(y.x)")
    assert_detected()


def test_attributes_3():
    run_cell(
        """
        class Foo:
            def __init__(self, x, y):
                self.x = x
                self.y = y
        """
    )
    run_cell("foo = Foo(5, 6)")
    run_cell("bar = Foo(7, 8)")
    run_cell("y = bar.x + 5")
    run_cell("foo.x = 8")
    run_cell("logging.info(y)")
    assert_not_detected("y does not depend on updated attrval foo.y")


def test_stale_use_of_attribute():
    run_cell(
        """
        class Foo:
            def __init__(self, x, y):
                self.x = x
                self.y = y
        """
    )
    run_cell("foo = Foo(5, 6)")
    run_cell("bar = Foo(7, 8)")
    run_cell("foo.x = bar.x + bar.y")
    run_cell("bar.y = 42")
    run_cell("logging.info(foo.x)")
    assert_detected("`foo.x` depends on stale `bar.y`")


def test_attr_manager_active_scope_resets():
    run_cell(
        """
        y = 10
        class Foo:
            def f(self):
                y = 11
                return y
        def f():
            return y
        """
    )
    run_cell("foo = Foo()")
    # if the active scope doesn't reset after done with foo.f(),
    # it will think the `y` referred to by f() is the one in Foo.f's scope.
    run_cell("x = foo.f() + f()")
    run_cell("y = 42")
    run_cell("logging.info(x)")
    assert_detected("`x` depends on stale `y`")


def test_attribute_call_point():
    run_cell(
        """
        y = 10
        class Foo:
            def f(self):
                return y
        """
    )
    run_cell("foo = Foo()")
    run_cell("x = foo.f()")
    run_cell("y = 42")
    run_cell("logging.info(x)")
    assert_detected("`x` depends on stale `y`")


def test_attr_manager_active_scope_with_property():
    run_cell(
        """
        y = 10
        class Foo:
            @property
            def f(self):
                y = 11
                return y
        """
    )
    run_cell("foo = Foo()")
    # if the active scope doesn't reset after done with foo.f(),
    # it will think the `y` referred to by f() is the one in Foo.f's scope.
    run_cell("x = foo.f")
    run_cell("y = 42")
    run_cell("logging.info(x)")
    assert_not_detected("`x` independent of outer `y`")


def test_namespace_scope_resolution():
    run_cell(
        """
        y = 42
        class Foo:
            y = 10
            @property
            def foo(self):
                return y
        """
    )
    run_cell("foo = Foo()")
    run_cell("x = foo.foo")
    run_cell("Foo.y = 99")
    run_cell("logging.info(x)")
    assert_not_detected("`x` should not have dependency on `Foo.y`")


def test_long_chain_attribute():
    run_cell(
        """
        class Foo:
            shared = 99
            def __init__(self, x, y):
                self.x = x
                self.y = y + self.shared
                
            class Bar:
                def __init__(self, a, b):
                    self.a = a
                    self.b = b
                    
                def foo(self):
                    return Foo(self.a, self.b)
                    
                def bar(self):
                    return Foo
        """
    )
    run_cell("foo = Foo(5, 6)")
    run_cell("bar = Foo.Bar(7, 8)")
    run_cell("foo.x = 42")
    run_cell("logging.info(bar.a)")
    assert_not_detected()
    run_cell("Foo.Bar(9, 10).foo().shared = 100")
    run_cell(
        """
        for _ in range(10**2):
            Foo.Bar(9, 10).foo().shared = 100
        """
    )
    run_cell("logging.info(foo.y)")
    assert_not_detected(
        "we mutated `shared` on obj and not on `Foo` (the one on which `foo.y` depends)"
    )
    run_cell("Foo.Bar(9, 10).bar().shared = 100")
    run_cell("logging.info(foo.y)")
    assert_detected("we mutated a shared value on which `foo.y` depends")


def test_numpy_subscripting():
    run_cell("import numpy as np")
    run_cell("x = np.zeros(5)")
    run_cell("y = x[3] + 5")
    run_cell("x[3] = 2")
    run_cell("logging.info(y)")
    assert_detected("y depends on stale x[3]")


def test_dict_subscripting():
    run_cell('d = {"foo": "bar", 0: "bat"}')
    run_cell("x = 7")
    run_cell("d[0] = x")
    run_cell("x += 1")
    run_cell('z = d["foo"] + " asdf"')
    assert_not_detected('`d["foo"]` does not depend on stale `x`, unlike `d[0]`')
    run_cell("logging.info(z)")
    assert_not_detected()
    run_cell('logging.info(d["foo"])')
    assert_not_detected()


def test_dict_subscripting_2():
    run_cell('d = {"foo": "bar", 0: "bat"}')
    run_cell("x = 7")
    run_cell("d[0] = x")
    run_cell("x += 1")
    run_cell("d[0] = x")
    run_cell("logging.info(d)")
    assert_not_detected("we updated the stale entry of `d`")


def test_pandas_attr_mutation_with_alias():
    run_cell("import pandas as pd")
    run_cell('df = pd.DataFrame({"a": [0,1], "b": [2., 3.]})')
    run_cell("asdf = df.b")
    run_cell("df.b *= 7")
    run_cell("logging.info(asdf)")
    assert_detected(
        "`asdf` has same values as `df.b`, but they now point to different things, "
        "which in general could be dangerous"
    )


def test_stale_detection_works_when_namespace_available_but_stale_symbol_unavailable():
    run_cell("import pandas as pd")
    run_cell('data = {"a": list(range(5)), "b": list(range(1, 6))}')
    run_cell("df = pd.DataFrame(data)")
    run_cell("df.b")  # force creation of a namespace for df
    run_cell('data = {"a": list(range(4)), "b": list(range(1, 5))}')
    run_cell("logging.info(df.a)")
    assert_detected("`df.a` depends on stale `data` dictionary")


def test_stale_detection_works_when_namespace_available_but_stale_symbol_unavailable_2():
    run_cell("import pandas as pd")
    run_cell('data = {"a": list(range(5)), "b": list(range(1, 6))}')
    run_cell("df = pd.DataFrame(data)")
    run_cell("df.b")  # force creation of a namespace for df
    run_cell('data["b"] = 42')
    run_cell("logging.info(df.a)")
    assert_false_positive('`df.a` independent of entry `data["b"]`')


def test_list_alias_breaking():
    run_cell("x = [0]")
    run_cell("y = x")
    run_cell("x = [7]")
    run_cell("logging.info(y)")
    assert_detected("`y` has stale dependency on old `x`")


def test_list_alias_no_break():
    run_cell("x = [0]")
    run_cell("y = x")
    run_cell("x *= 7")
    run_cell("logging.info(y)")
    assert_not_detected("`y` still aliases `x`")


@skipif_known_failing
def test_subscript_sensitivity():
    run_cell("lst = list(range(5))")
    run_cell("i = 0")
    run_cell("lst[i] = 10")
    run_cell("i = 1")
    run_cell("logging.info(lst)")
    assert_detected("`lst[0]` depends on stale i")


def test_subscript_adds():
    run_cell("x = 42")
    run_cell(
        """
        d = {
            'foo': x,
            'bar': 77
        }
        """
    )
    run_cell('d["bar"] = 99')
    run_cell("x = 100")
    run_cell("logging.info(d)")
    assert_detected("`d` depends on stale `x`")


def test_list_mutation():
    run_cell("lst = list(range(5))")
    run_cell("x = 42")
    run_cell("asdf = []")
    run_cell("asdf.append(lst.append(x))")
    run_cell("x = 43")
    run_cell("logging.info(lst)")
    assert_detected("lst depends on stale x")


def test_list_mutation_extend():
    run_cell("lst = [0, 1]")
    run_cell("x = lst[0] + 7")
    run_cell("lst.extend([2, 3, 4, x])")
    run_cell("logging.info(x)")
    assert_not_detected()
    run_cell("x = 77")
    run_cell("logging.info(lst)")
    assert_detected("`lst` depends on stale `x`")


def test_list_mutation_2():
    run_cell("lst = list(range(5))")
    run_cell("x = lst + [42, 43]")
    run_cell("lst.append(99)")
    run_cell("logging.info(x)")
    assert_detected("`x` depends on stale `lst`")


def test_list_mutation_from_attr():
    run_cell('s = "hello X world X how X are X you X today?"')
    run_cell("lst = []")
    run_cell(
        """
        for word in s.split('X'):
            lst.append(word.strip())
        """
    )
    run_cell('s = "foobar"')
    run_cell("logging.info(lst)")
    assert_detected("`lst` depends on stale `s`")


@skipif_known_failing
def test_list_mutation_extend_from_attr():
    # for this one, we somehow have to avoid disabling tracing which is hard
    run_cell('s = "hello X world X how X are X you X today?"')
    run_cell("lst = []")
    run_cell('lst.extend(word.strip() for word in s.split("X"))')
    run_cell('s = "foobar"')
    run_cell("logging.info(lst)")
    assert_detected("`lst` depends on stale `s`")


def test_lazy_class_scope_resolution():
    run_cell(
        """
        class Foo:
            shared = 99
            def __init__(self, x):
                self.x = x
        """
    )
    run_cell("foo = Foo(10)")
    run_cell("y = 11")
    run_cell("Foo.shared = y + 42")
    run_cell("y = 12")
    run_cell("logging.info(foo.shared)")
    assert_detected(
        "`foo.shared` should point to same DataSymbol as `Foo.shared` and thus also has stale dep"
    )
    run_cell("foo.shared = 89")
    run_cell("logging.info(Foo.shared)")
    assert_detected(
        "Even though we refreshed `foo.shared`, this "
        "has no bearing on the original class member `Foo.shared`"
    )


def test_new_scope_val_depends_on_old():
    run_cell(
        """
        class Foo:
            shared = 99
        """
    )
    run_cell("foo = Foo()")
    run_cell("foo.shared = 11")
    run_cell("foo_shared_alias = foo.shared")
    run_cell("Foo.shared = 12")
    run_cell("logging.info(foo_shared_alias)")
    assert_detected()
    run_cell("logging.info(foo.shared)")
    assert_detected()
    run_cell("logging.info(foo)")
    assert_detected()


def test_class_member_mutation_does_not_affect_instance_members():
    run_cell(
        """
        class Foo:
            shared = 99
            def __init__(self):
                self.x = 42
        """
    )
    run_cell("foo = Foo()")
    run_cell("Foo.shared = 12")
    run_cell("logging.info(foo.x)")
    assert_not_detected()


def test_numpy_subscripting_fp():
    run_cell("import numpy as np")
    run_cell("x = np.zeros(5)")
    run_cell("y = x[3] + 5")
    run_cell("x[0] = 2")
    run_cell("logging.info(y)")
    assert_not_detected("`y` depends on unchanged `x[3]` and not on changed `x[0]`")


def test_old_format_string():
    run_cell("a = 5\nb = 7")
    run_cell('expr_str = "{} + {} = {}".format(a, b, a + b)')
    run_cell("a = 9")
    run_cell("logging.info(expr_str)")
    assert_detected("`expr_str` depends on stale `a`")


def test_old_format_string_kwargs():
    run_cell("a = 5\nb = 7")
    run_cell('expr_str = "{a} + {b} = {total}".format(a=a, b=b, total=a + b)')
    run_cell("a = 9")
    run_cell("logging.info(expr_str)")
    assert_detected("`expr_str` depends on stale `a`")


def test_new_format_string():
    run_cell("a = 5\nb = 7")
    run_cell('expr_str = f"{a} + {b} = {a+b}"')
    run_cell("a = 9")
    run_cell("logging.info(expr_str)")
    assert_detected("`expr_str` depends on stale `a`")


def test_scope_resolution():
    run_cell(
        """
        def f(x):
            def g(x):
                return 2 * x
            return g(x) + 8
        """
    )
    run_cell("x = 7")
    run_cell("y = f(x)")
    run_cell("x = 8")
    run_cell("logging.info(y)")
    assert_detected("`y` depends on stale `x`")


def test_scope_resolution_2():
    run_cell(
        """
        def g(x):
            return 2 * x
        def f(x):
            return g(x) + 8
        """
    )
    run_cell("x = 7")
    run_cell("y = f(x)")
    run_cell("x = 8")
    run_cell("logging.info(y)")
    assert_detected("`y` depends on stale `x`")


def test_funcall_kwarg():
    run_cell(
        """
        def f(y):
            return 2 * y + 8
        """
    )
    run_cell("x = 7")
    run_cell("z = f(y=x)")
    run_cell("x = 8")
    run_cell("logging.info(z)")
    assert_detected("`z` depends on stale `x`")


def test_funcall_kwarg_2():
    run_cell(
        """
        def f(y):
            return 2 * y + 8
        """
    )
    run_cell("x = 7")
    run_cell("y = f(y=x)")
    run_cell("x = 8")
    run_cell("logging.info(y)")
    assert_detected("`y` depends on stale `x`")


def test_funcall_kwarg_3():
    run_cell(
        """
        def f(x):
            return 2 * x + 8
        """
    )
    run_cell("x = 7")
    run_cell("y = f(x=x)")
    run_cell("x = 8")
    run_cell("logging.info(y)")
    assert_detected("`y` depends on stale `x`")


def test_funcall_kwarg_4():
    run_cell(
        """
        def f(x):
            return 2 * x + 8
        """
    )
    run_cell("x = 7")
    run_cell("x = f(x=x)")
    run_cell("x = 8")
    run_cell("logging.info(x)")
    assert_not_detected("`x` is overriden so should not be stale")


def test_single_line_dictionary_literal():
    run_cell("foo = 5")
    run_cell("bar = 6")
    run_cell('d = {foo: bar, "pi": 42,}')
    run_cell("bar = 7")
    run_cell("logging.info(d)")
    assert_detected("`d` depends on stale `bar`")


def test_single_line_dictionary_literal_fix_stale_deps():
    run_cell("foo = 5")
    run_cell("bar = 6")
    run_cell('d = {foo: bar, "pi": 42,}')
    run_cell("bar = 7")
    run_cell("logging.info(d)")
    assert_detected("`d` depends on stale `bar`")
    run_cell("d[foo] = bar")
    assert_not_detected()
    run_cell("logging.info(d)")
    # assert_false_positive('`d`s stale dep fixed, but this is hard to detect '
    #                       'since we did not yet have a DataSymbol for `d[foo]` when staleness introduced')
    assert_not_detected("`d`s stale dep fixed")
    run_cell("foo = 8")
    run_cell("logging.info(d)")
    assert_detected("`d` depends on stale `foo`")
    # TODO: not sure what the correct behavior for the below write to d[foo] after foo changed should be
    # run_cell('d[foo] = bar')
    # run_cell('logging.info(d)')
    # assert_not_detected('`d`s stale dep fixed')


def test_multiline_dictionary_literal():
    run_cell("foo = 5")
    run_cell("bar = 6")
    run_cell(
        """
        d = {
            foo: bar,
            'pi': 42,
        }
        """
    )
    run_cell("bar = 7")
    run_cell("logging.info(d)")
    assert_detected("`d` depends on stale `bar`")


def test_exception():
    run_cell("lst = list(range(5))")
    run_cell("x = 6")
    run_cell(
        """
        try:
            lst[x] = 42
        except:
            lst[0] = 42
        """
    )
    run_cell("x = 7")
    run_cell("logging.info(lst)")
    assert_not_detected("lst should be independent of x due to exception")


def test_for_loop_binding():
    run_cell("a = 0")
    run_cell("b = 1")
    run_cell("c = 2")
    run_cell("lst = [a, b, c]")
    run_cell(
        """
        for i in lst:
            pass
        """
    )
    run_cell("a = 3")
    run_cell("logging.info(i)")
    assert_false_positive(
        "`i` should not depend on `a` at end of for loop but this is hard"
    )


@skipif_known_failing
def test_for_loop_literal_binding():
    run_cell("a = 0")
    run_cell("b = 1")
    run_cell("c = 2")
    run_cell(
        """
        for i in [a, b, c]:
            pass
        """
    )
    run_cell("a = 3")
    run_cell("logging.info(i)")
    assert_not_detected("`i` should not depend on `a` at end of for loop")


def test_for_loop_partial_dep():
    run_cell("lst = list(range(10))")
    run_cell("s = 0")
    run_cell(
        """
        for i in range(5):
            s += lst[i]
        """
    )
    run_cell("lst[-1] = 42")
    run_cell("logging.info(s)")
    assert_not_detected("`s` does not depend on last entry of `lst`")
    run_cell("lst[1] = 22")
    run_cell("logging.info(s)")
    assert_false_negative(
        "`s` does depend on second entry of `lst` but tracing every iteration of loop is slow"
    )


def test_for_loop_tuple_unpack():
    run_cell("x = (1, 2)")
    run_cell("y = (3, 4)")
    run_cell("total_i = 0")
    run_cell("total_j = 0")
    run_cell(
        """
        for i, j in [x, y]:
            total_i += i
            total_j += j
        """
    )
    run_cell("logging.info(total_i)")
    assert_not_detected()
    run_cell("logging.info(total_j)")
    assert_not_detected()
    run_cell("x = (42, 43)")
    run_cell("logging.info(total_i)")
    assert_detected("`total_i` depends on old version of `x`")
    run_cell("logging.info(total_j)")
    assert_detected("`total_j` also depends on old version of `x`")


def test_same_cell_redefine():
    run_cell("a = 0")
    run_cell(
        """
        b = a + 1
        a = 42
        """
    )
    run_cell("logging.info(b)")
    assert_not_detected(
        "`b` should not be considered as having stale dependency since `a` changed in same cell as `b`"
    )


def test_multiple_stmts_in_one_line():
    run_cell("a = 1; b = 2")
    run_cell("x = a + b")
    run_cell("a = 42")
    run_cell("logging.info(x)")
    assert_detected("`x` depends on stale value of `a`")


def test_multiple_stmts_in_one_line_2():
    run_cell("a = 1; b = 2")
    run_cell("x = a + b")
    run_cell("b = 42")
    run_cell("logging.info(x)")
    assert_detected("`x` depends on stale value of `a`")


def test_line_magic():
    run_cell(
        """
        %lsmagic
        %lsmagic
        %lsmagic
        %lsmagic
        %lsmagic
        a = 0
        """
    )
    run_cell(
        """
        %lsmagic
        %lsmagic
        %lsmagic
        %lsmagic
        x = a + 1
        """
    )
    run_cell(
        """
        %lsmagic
        %lsmagic
        %lsmagic
        a = 42
        """
    )
    run_cell(
        """
        %lsmagic
        %lsmagic
        logging.info(x)
        """
    )
    assert_detected("`x` depends on stale value of `a`")
    run_cell(
        """
        %lsmagic
        logging.info(x)
        %lsmagic
        """
    )
    assert_detected("`x` depends on stale value of `a`")


def test_exception_stack_unwind():
    def assert_stack_size(size):
        len_call_stack = "len(tracer().call_stack)"
        return ", ".join(
            [
                f"assert {len_call_stack} == {size}",
                f'"%d vs {size}" % {len_call_stack}',
            ]
        )

    run_cell(
        f"""
        import numpy as np
        from nbsafety.singletons import tracer
        {assert_stack_size(0)}
        def f():
            {assert_stack_size(1)}
            def g():
                {assert_stack_size(2)}
                def h():
                    {assert_stack_size(3)}
                    return np.loadtxt('does-not-exist.txt')
                return h()
            try:
                return g()
            except:
                {assert_stack_size(1)}
        f()
        {assert_stack_size(0)}
        """
    )


def test_throwing_statements_do_not_track_deps():
    run_cell(
        """
        z = 10
        def foo():
            def bar():
                raise ValueError('foo!')
            return bar() + z
        x = 0
        y = x + 1
        """
    )
    run_cell(
        """
        try:
            x = 42 + foo()
        except:
            pass
        """
    )
    run_cell("logging.info(y)")
    assert_not_detected("no stale dep for `y` because update on `x` threw exception")
    run_cell("z = 99")
    run_cell("logging.info(x)")
    assert_not_detected(
        "no stale dep for `x` because it is indep of `z` (attempted dep add threw)"
    )


def test_attr_dep_from_somewhere_else():
    run_cell("import fakelib as fake")
    run_cell("fake.y = 7")
    run_cell("x = fake.y + 1")
    run_cell("fake.y = 42")
    run_cell("logging.info(x)")
    assert_detected("`x` depends on old value of `fake.y`")


def test_attr_use_from_somewhere_else():
    run_cell("import fakelib as fake")
    run_cell("x = 7")
    run_cell("fake.y = x + 1")
    run_cell("x = 42")
    run_cell("logging.info(fake.y)")
    assert_detected("`fake.y` depends on old value of `x`")


def test_class_assignment():
    run_cell(
        """
        class Foo:
            def __init__(self):
                self.y = 99

        Bar = Foo
        foo = Bar()
        x = 7
        """
    )
    run_cell("foo.y = x + 1")
    run_cell("x = 42")
    run_cell("logging.info(foo.y)")
    assert_detected("`foo.y` depends on stale `x`")
    run_cell("foo.y = 10")
    run_cell("x = foo.y + 1")
    run_cell("foo.y = 12")
    run_cell("logging.info(x)")
    assert_detected("`x` depends on stale `foo.y`")


def test_no_class_false_positives():
    run_cell("x = 7")
    run_cell("y = x + 1")
    run_cell(
        """
        class Foo:
            x = 42
        """
    )
    run_cell("logging.info(y)")
    assert_not_detected("x inside class scope is different")


def test_set_literal():
    run_cell("x, y, z = 1, 2, 3")
    run_cell("s = {x + 1, y + 7}")
    run_cell("z = 42")
    run_cell("logging.info(s)")
    assert_not_detected()
    run_cell("x = 17")
    run_cell("logging.info(s)")
    assert_detected()


def test_tuple_unpack_simple():
    run_cell("x, y = 0, 1")
    run_cell("a, b = x + 2, y + 3")
    run_cell("x, y = 42, 43")
    run_cell("logging.info(a)")
    assert_detected("`a` depends on stale `x`")
    run_cell("logging.info(b)")
    assert_detected("`b` depends on stale `y`")


def test_tuple_unpack_hard():
    run_cell("x, y = 0, 1")
    run_cell("a, b = x + 2, y + 3")
    run_cell("y = 43")
    run_cell("logging.info(a)")
    assert_not_detected("`a` does not depend on `y`")
    run_cell("logging.info(b)")
    assert_detected("`b` depends on stale `y`")
    run_cell("b = y + 10")
    run_cell("x = 99")
    run_cell("logging.info(b)")
    assert_not_detected("`b` does not depend on `x`")
    run_cell("logging.info(a)")
    assert_detected("`a` depends on stale `x`")


def test_unpack_multiple_from_single():
    run_cell("x, y = 0, 1")
    run_cell("lst = [x + 1, y + 1]")
    run_cell("a, b = lst")
    run_cell("x = 42")
    run_cell("logging.info(b)")
    assert_not_detected()
    run_cell("logging.info(a)")
    assert_detected()


@skipif_known_failing
def test_attr_dep_with_top_level_overwrite():
    run_cell(
        """
        class Foo:
            def __init__(self):
                self.y = 99
        foo = Foo()
        x = 42
        foo.y = x + 7
        """
    )
    run_cell("x = 43")
    run_cell("logging.info(foo)")  # this should be a 'deep' usage of foo
    assert_detected("logging.info could display `foo.y` which depends on x")
    run_cell("foo.y = 70")
    assert_not_detected("we just fixed stale dep of `foo.y` by changing its deps")
    run_cell("x = foo.y + 7")
    run_cell("foo = 81")
    run_cell("logging.info(x)")
    assert_detected("`x` has stale dep on `foo` (transitively through `foo.y`)")


def test_typed_assignment():
    run_cell("a = 0")
    run_cell("b: int = a + 1")
    run_cell("a: int = 42")
    run_cell("logging.info(b)")
    assert_detected("`b` has stale dep on `a`")


@skipif_known_failing
def test_time_line_magic():
    run_cell("a = 0")
    run_cell("b = a + 1")
    run_cell("%time a = 42")
    run_cell("logging.info(b)")
    assert_detected("`b` has stale dep on `a`")


@skipif_known_failing
def test_cell_magic():
    run_cell("a = 0")
    run_cell("b = a + 1")
    run_cell("%%time\na = 42")
    run_cell("logging.info(b)")
    assert_detected("`b` has stale dep on `a`")


def test_pandas():
    run_cell("import numpy as np")
    run_cell("import pandas as pd")
    run_cell("arr = 1 + np.arange(10)")
    run_cell('df = pd.DataFrame({"col": arr})')
    run_cell("df2 = df.dropna()")
    run_cell("df.dropna()")
    run_cell("logging.info(df2)")
    assert_not_detected("`df.dropna()` did not mutate `df`")
    run_cell("df.dropna(inplace=True)")
    run_cell("logging.info(df2)")
    assert_detected("`df.dropna(inplace=True)` mutated `df`")


def test_deeply_nested_arg_and_kwarg_refs_with_attr_calls():
    run_cell(
        """
        class Foo:
            def foo(self, x, y=0):
                return x + y
        """
    )
    run_cell("x = 1")
    run_cell("y = 2")
    run_cell("foo = Foo()")
    run_cell("z = foo.foo(foo.foo(foo.foo(x, y=y), y=7), y=foo.foo(123))")
    run_cell("logging.info(z)")
    assert_not_detected()
    run_cell("x = 101")
    run_cell("logging.info(z)")
    assert_detected("`z` depends on stale `x`")
    run_cell("z = foo.foo(foo.foo(foo.foo(x, y=y), y=foo.foo(200)), y=99)")
    run_cell("logging.info(z)")
    assert_not_detected()
    run_cell("y = 102")
    run_cell("logging.info(z)")
    assert_detected("`z` depends on stale `y`")


def test_class_member_granularity():
    run_cell(
        """
        y = 77
        class Foo:
            def __init__(self):
                self.x = 10
                self.y = y
        """
    )
    run_cell("foo = Foo()")
    run_cell("z = foo.x + 7")
    run_cell("y = 42")
    run_cell("logging.info(z)")
    assert_not_detected("`z` independent of stale `y`")


def test_no_rhs_propagation():
    run_cell("import numpy as np")
    run_cell("x = np.random.random(10)")
    run_cell("y = np.random.random(10)")
    run_cell("inds = np.argsort(x)")
    run_cell("x = x[inds]")
    run_cell("y = y[inds]")
    assert_not_detected(
        "`inds` should not be considered stale since it appears on RHS of assignment"
    )


def test_if_true():
    run_cell("y = 0")
    run_cell("z = 42")
    run_cell(
        """
        if True:
            x = y + 1
        else:
            x = z + 1
        """
    )
    run_cell("z = 43")
    run_cell("logging.info(x)")
    assert_not_detected("`x` not dependent on `z`")
    run_cell("y = 99")
    run_cell("logging.info(x)")
    assert_detected("`x` dependent on old `y`")


def test_tracing_reenabled_after_dup_funcall():
    run_cell("x = 0")
    run_cell(
        """
        def f():
            return 42
        """
    )
    run_cell(
        """
        f()
        f()
        y = x + 1
        """
    )
    run_cell("x = 42")
    run_cell("logging.info(y)")
    assert_detected()


def test_one_time_tracing_func():
    run_cell("x = 0")
    run_cell("y = 1")
    run_cell(
        """
        def f(p):
            if p:
                return x
            else:
                return y
        """
    )
    run_cell("z = f(False) + 1\nz = f(True) + 1")
    run_cell("y = 2")
    run_cell("logging.info(z)")
    assert_not_detected()
    run_cell("x = 3")
    run_cell("logging.info(z)")
    assert_detected("tracing should not be disabled")


def test_tracing_disable_with_nested_calls():
    # run_cell('%safety trace_messages enable')
    run_cell("y = 0")
    run_cell(
        """
        def f():
            return y
        """
    )
    run_cell(
        """
        def g(flag):
            if flag:
                return f()
            else:
                return 2
        """
    )
    run_cell(
        """
        g(False)
        x = g(True) + 1
        """
    )
    run_cell("y = 42")
    run_cell("logging.info(x)")
    assert_detected("`x` has dep on stale `y`")


def test_dict():
    run_cell("d = {}; d[0] = 0")
    run_cell("x = d[0] + 1")
    run_cell("d[0] = 42")
    run_cell("logging.info(x)")
    assert_detected("`x` has dependency on old value of `d[0]`")


def test_dict_2():
    run_cell("d = {}; d[0] = 0")
    run_cell("x = d[0] + 1")
    run_cell("d = {}; d[0] = 42")
    run_cell("logging.info(x)")
    assert_detected("`x` has dependency on old value of `d[0]`")


def test_default_dict():
    run_cell("from collections import defaultdict")
    run_cell("d = defaultdict(dict); d[0][0] = 0")
    run_cell("x = d[0][0] + 1")
    run_cell("d = defaultdict(dict); d[0][0] = 42")
    run_cell("logging.info(x)")
    assert_detected("`x` has dependency on old value of `d[0][0]`")


def test_mutate_arg():
    run_cell("import numpy as np")
    run_cell("x = np.ones(5)")
    run_cell("y = x + 1")
    run_cell("np.random.shuffle(x)")
    run_cell("logging.info(y)")
    assert_detected("`y` has a dependency on an old value of `x`")


def test_mutate_arg_special_cases():
    run_cell("import numpy as np")
    run_cell("x = np.random.random(10)")
    run_cell("y = np.ones(5)")
    run_cell("np.random.seed(42)")
    run_cell("logging.info(x)")
    assert_detected("`x` depends on previous seed")
    run_cell("logging.info(y)")
    assert_not_detected("`y` has no dependency on seed")
    run_cell("x = y + 5")
    run_cell("logging.info(y)")
    assert_not_detected()
    run_cell("logging.info(x)")
    assert_not_detected("`logging.info` does not mutate `y`")


def test_augassign_does_not_overwrite_deps():
    run_cell("x = 0")
    run_cell("y = 1")
    run_cell("z = x + 2")
    run_cell("z += y")
    run_cell("x = 42")
    run_cell("logging.info(z)")
    assert_detected("`z` depends on old value of `x`")


def test_rval_appearance_does_not_overwrite_deps():
    run_cell("x = 0")
    run_cell("y = 1")
    run_cell("z = x + 2")
    run_cell("z = z + y")
    run_cell("x = 42")
    run_cell("logging.info(z)")
    assert_detected("`z` depends on old value of `x`")


def test_augassign_does_not_overwrite_deps_for_attributes():
    run_cell(
        """
        class Foo:
            def __init__(self, x):
                self.x = x
        """
    )
    run_cell("x, y, z = Foo(0), Foo(5), Foo(10)")
    run_cell("z.x = x.x + 2")
    run_cell("z.x += y.x")
    run_cell("x.x = 42")
    run_cell("logging.info(z.x)")
    assert_detected("`z.x` depends on old value of `x.x`")


def test_rval_appearance_does_not_overwrite_deps_for_attributes():
    run_cell(
        """
        class Foo:
            def __init__(self, x):
                self.x = x
        """
    )
    run_cell("x, y, z = Foo(0), Foo(5), Foo(10)")
    run_cell("z.x = x.x + 2")
    run_cell("z.x = z.x + y.x")
    run_cell("x.x = 42")
    run_cell("logging.info(z.x)")
    assert_detected("`z.x` depends on old value of `x.x`")


def test_context_manager():
    run_cell(
        """
        from contextlib import contextmanager

        @contextmanager
        def foo():
            yield 42
        """
    )
    run_cell("with foo() as bar: x = bar + 7")
    run_cell("bar = 43")
    run_cell("logging.info(x)")
    assert_detected("`x` depends on old value of `bar`")


def test_decorator():
    run_cell("foo = lambda f: f")
    run_cell(
        """
        @foo
        def bar(x):
            return x + 42
        """
    )
    run_cell("y = bar(7)")
    run_cell("foo = lambda f: lambda x: f(x + 9)")
    run_cell("logging.info(y)")
    assert_detected("`y` depends on call to `bar` which has stale decorator `@foo`")


def test_magics_dont_break_things():
    run_cell(
        """
        %time dummy = 0
        x = 0
        y = x + 1
        """
    )
    run_cell("x = 42")
    run_cell("logging.info(y)")
    assert_detected("`y` depends on old value `x`")


def test_tuple_return():
    run_cell(
        """
        x = 0
        y = 1
        a = x + 42
        b = y + 77
        def foo():
            return a, b
        """
    )
    run_cell("t = foo()[1]")
    run_cell("x = 9")
    run_cell("logging.info(t)")
    assert_not_detected("`t` independent of updated `x`")
    run_cell("y = 10")
    run_cell("logging.info(t)")
    assert_detected("`t` depends on old version of `y`")


def test_tuple_return_obj():
    run_cell(
        """
        x = 0
        y = 1
        a = x + 42
        b = y + 77
        def foo():
            return [a], [b]
        """
    )
    run_cell("t = foo()[1][0]")
    run_cell("x = 9")
    run_cell("logging.info(t)")
    assert_not_detected("`t` independent of updated `x`")
    run_cell("y = 10")
    run_cell("logging.info(t)")
    assert_detected("`t` depends on old version of `y`")


def test_property_in_function_arg():
    run_cell(
        f"""
        import builtins
        z = 42
        class Foo:
            @property
            def bar(self):
                # ensures that this is not called during static analysis
                assert hasattr(builtins, "{EMIT_EVENT}")
                return z

        def f(x, y):
            return x + y
        """
    )
    run_cell("foo = Foo()")
    run_cell("w = f(3, foo.bar)")
    run_cell("z = 7")
    run_cell("logging.info(w)")
    assert_detected("`w` depends on old version of `z`")


def test_getter_setter_with_global():
    run_cell(
        """
        z = 42
        class Bar:
            shared = 25
            @property
            def baz(self):
                return z
            
            @baz.setter
            def baz(self, new_val):
                global z
                z = new_val
            
        class Foo:
            def __init__(self):
                self.bar = Bar()

        def f(x, y):
            return x + y
        """
    )
    run_cell("foo = Bar()")
    run_cell("w = f(3, foo.baz)")
    run_cell("logging.info(w)")
    assert_not_detected()
    run_cell("foo.baz = 84")
    run_cell("logging.info(w)")
    assert_detected("`w` depends on stale `foo.baz`")


def test_list_extend():
    run_cell("lst = [0, 1, 2]")
    run_cell("x = lst[1] + 1")
    run_cell("lst.extend([3, 4, 5, 6, 7, 8, 9])")
    run_cell("logging.info(x)")
    assert_not_detected()
    run_cell("logging.info(lst[1])")
    assert_not_detected()
    run_cell("y = lst[8] + 1")
    run_cell("lst[8] += 42")
    run_cell("logging.info(y)")
    assert_detected("`y` depends on stale `lst[8]`")


@skipif_known_failing
def test_list_sum_simple():
    run_cell("w, x, y, z = 0, 1, 2, 3")
    run_cell("lst = [w, x] + [y, z]")
    run_cell("z += 42")
    run_cell("logging.info(lst[0])")
    assert_not_detected()
    run_cell("logging.info(lst[1])")
    assert_not_detected()
    run_cell("logging.info(lst[2])")
    assert_not_detected()
    run_cell("logging.info(lst[3])")
    assert_detected()


def test_global_var():
    run_cell("x = 0")
    run_cell("y = x + 1")
    run_cell("def f(): global x; x = 42")
    run_cell("logging.info(y)")
    assert_not_detected()
    run_cell("f()")
    run_cell("logging.info(y)")
    assert_detected()


def test_nonlocal_var():
    cell_template = """
        z = 0
        def f():
            x = 0
            def g():
                {stmt}
                x = 42
            global z
            z = x
            return g
        g = f()
        """
    run_cell(cell_template.format(stmt="pass"))
    run_cell("y = z + 1")
    run_cell("g()")
    run_cell("logging.info(y)")
    assert_not_detected()
    run_cell(cell_template.format(stmt="nonlocal x"))
    run_cell("y = z + 1")
    run_cell("g()")
    run_cell("logging.info(y)")
    assert_detected()


def test_reimport_does_not_propagate():
    run_cell("import numpy as np")
    run_cell("arr = np.zeros(3)")
    run_cell("import numpy as np")
    assert_not_detected()


def test_pyccolo_exec():
    run_cell("x = 0")
    run_cell("y = x + 1")
    run_cell("import pyccolo as pyc; _ = pyc.exec('x = 42')")
    run_cell("logging.info(y)")
    assert_not_detected()


def test_pyccolo_exec_2():
    run_cell("lst = [42]")
    run_cell("lst2 = lst + [9001]")
    run_cell("import pyccolo as pyc; _ = pyc.exec('lst.append(55)')")
    run_cell("logging.info(lst2)")
    assert_not_detected()
    run_cell("assert lst == [42, 55]")


# TODO: where was I going with this?
# def test_getitem_call():
#     run_cell("""
# class Foo:
#     def __getitem__(self, x):
#         return 42
# """)
#     run_cell('foo = Foo()')
#     run_cell('x = foo[0]')


if sys.version_info >= (3, 8):

    def test_walrus_simple():
        run_cell(
            """
            if (x := 1) > 0:
                y = x + 1
            """
        )
        run_cell("x = 42")
        run_cell("logging.info(y)")
        assert_detected("`y` depends on old value of `x`")

    def test_walrus_fancy():
        run_cell(
            """
            if (x := (y := (z := 1) + 1) + 1) > 0:
                a = x + 1
            """
        )
        run_cell("y = 42")
        run_cell("logging.info(z)")
        assert_not_detected("`z` does not depend on `y`")
        run_cell("logging.info(y)")
        assert_not_detected("`y` is updated")
        run_cell("logging.info(x)")
        assert_detected("`x` depends on old value of `y`")
        run_cell("logging.info(a)")
        assert_detected("`a` depends on old value of `y`")

    def test_walrus_fancy_attributes():
        run_cell(
            """
            class Foo:
                def __init__(self, x):
                    self.x = x
            """
        )
        run_cell("foo = Foo(9001)")
        run_cell(
            """
            if (x := (y := (z := 1) + foo.x) + 1) > 0:
                a = x + 1
            """
        )
        run_cell("foo.x = 42")
        run_cell("logging.info(z)")
        assert_not_detected("`z` does not depend on `foo.x`")
        run_cell("logging.info(y)")
        assert_detected("`y` depends on old `foo.x`")
        run_cell("logging.info(x)")
        assert_detected("`x` depends on old value of `foo.x`")
        run_cell("logging.info(a)")
        assert_detected("`a` depends on old value of `foo.x`")



# -*- coding: utf-8 -*-
import logging
from typing import TYPE_CHECKING, Generator, Iterable, Set

from nbsafety.data_model.timestamp import Timestamp
from nbsafety.singletons import nbs, tracer

if TYPE_CHECKING:
    # avoid circular imports
    from nbsafety.data_model.data_symbol import DataSymbol

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


class UpdateProtocol:
    def __init__(self, updated_sym: "DataSymbol") -> None:
        self.updated_sym = updated_sym
        self.seen: Set["DataSymbol"] = set()

    def __call__(
        self,
        new_deps: Set["DataSymbol"],
        mutated: bool,
        propagate_to_namespace_descendents: bool,
        refresh: bool,
    ) -> None:
        # in most cases, mutated implies that we should propagate to namespace descendents, since we
        # do not know how the mutation affects the namespace members. The exception is for specific
        # known events such as 'list.append()' or 'list.extend()' since we know these do not update
        # the namespace members.
        logger.warning(
            "updated sym %s (containing scope %s) with children %s",
            self.updated_sym,
            self.updated_sym.containing_scope,
            self.updated_sym.children,
        )
        directly_updated_symbols = (
            nbs().aliases[self.updated_sym.obj_id] if mutated else {self.updated_sym}
        )
        directly_updated_symbols |= self._maybe_get_adhoc_pandas_updated_syms()
        self._collect_updated_symbols_and_refresh_namespaces(
            directly_updated_symbols, propagate_to_namespace_descendents
        )
        logger.warning(
            "for symbol %s: mutated=%s; updated_symbols=%s",
            self.updated_sym,
            mutated,
            directly_updated_symbols,
        )
        updated_symbols_with_ancestors = set(self.seen)
        logger.warning(
            "all updated symbols for symbol %s: %s",
            self.updated_sym,
            updated_symbols_with_ancestors,
        )
        tracer().this_stmt_updated_symbols |= self.seen
        if refresh:
            for updated_sym in directly_updated_symbols:
                if not updated_sym.is_stale and updated_sym is not self.updated_sym:
                    updated_sym.refresh()
        self.seen |= new_deps  # don't propagate to stuff on RHS
        for dsym in updated_symbols_with_ancestors:
            self._propagate_staleness_to_deps(dsym, skip_seen_check=True)

    def _maybe_get_adhoc_pandas_updated_syms(self):
        try:
            import pandas
        except ImportError:
            return set()

        if self.updated_sym.obj is None or not isinstance(
            self.updated_sym.obj, pandas.Series
        ):
            return set()

        ns = self.updated_sym.containing_namespace
        if ns is None or ns.obj is None or not isinstance(ns.obj, pandas.DataFrame):
            return set()

        name = self.updated_sym.name
        return {
            ns.lookup_data_symbol_by_name_this_indentation(name, is_subscript=is_sub)
            for is_sub in [True, False]
        } - {None}

    def _collect_updated_symbols_and_refresh_namespaces(
        self,
        updated_symbols: Iterable["DataSymbol"],
        refresh_descendent_namespaces: bool,
    ) -> None:
        logger.warning(
            "collecting updated symbols and namespaces for %s", updated_symbols
        )
        for dsym in updated_symbols:
            if dsym.is_import or dsym in self.seen:
                continue
            dsym.updated_timestamps.add(Timestamp.current())
            self.seen.add(dsym)
            for cell in dsym.cells_where_deep_live:
                cell.add_used_cell_counter(dsym, nbs().cell_counter())
            containing_ns = dsym.containing_namespace
            if containing_ns is not None:
                logger.warning(
                    "containing scope for %s: %s; ids %s, %s",
                    dsym,
                    containing_ns,
                    dsym.obj_id,
                    containing_ns.obj_id,
                )
                containing_ns.namespace_stale_symbols.discard(dsym)
                containing_ns.max_descendent_timestamp = Timestamp.current()
                self._collect_updated_symbols_and_refresh_namespaces(
                    nbs().aliases[containing_ns.obj_id], refresh_descendent_namespaces
                )
            if refresh_descendent_namespaces:
                dsym_ns = dsym.namespace
                if dsym_ns is not None:
                    self._collect_updated_symbols_and_refresh_namespaces(
                        dsym_ns.all_data_symbols_this_indentation(),
                        refresh_descendent_namespaces,
                    )

    def _propagate_staleness_to_namespace_parents(
        self, dsym: "DataSymbol", skip_seen_check: bool = False
    ) -> None:
        if not skip_seen_check and dsym in self.seen:
            return
        self.seen.add(dsym)
        containing_ns = dsym.containing_namespace
        if containing_ns is None:
            return
        logger.warning("add %s to namespace stale symbols of %s", dsym, containing_ns)
        containing_ns.namespace_stale_symbols.add(dsym)
        for containing_alias in nbs().aliases[containing_ns.obj_id]:
            self._propagate_staleness_to_namespace_parents(containing_alias)

        for containing_alias in nbs().aliases[containing_ns.obj_id]:
            # do this in 2 separate loops to make sure all containing_alias are added to 'seen'
            # works around the issue when one alias depends on another
            for child in self._non_class_to_instance_children(containing_alias):
                logger.warning(
                    "propagate from namespace parent of %s to child %s", dsym, child
                )
                self._propagate_staleness_to_deps(child)

    def _non_class_to_instance_children(
        self, dsym: "DataSymbol"
    ) -> Generator["DataSymbol", None, None]:
        if self.updated_sym is dsym:
            yield from dsym.children
            return
        for child in dsym.children:
            # Next, complicated check to avoid propagating along a class -> instance edge.
            # The only time this is OK is when we changed the class, which will not be the case here.
            child_namespace = child.namespace
            if child_namespace is not None and child_namespace.cloned_from is not None:
                if child_namespace.cloned_from.obj_id == dsym.obj_id:
                    continue
            yield child

    def _propagate_staleness_to_namespace_children(
        self, dsym: "DataSymbol", skip_seen_check: bool = False
    ) -> None:
        if not skip_seen_check and dsym in self.seen:
            return
        self.seen.add(dsym)
        self_ns = nbs().namespaces.get(dsym.obj_id, None)
        if self_ns is None:
            return
        for ns_child in self_ns.all_data_symbols_this_indentation(exclude_class=True):
            logger.warning("propagate from %s to namespace child %s", dsym, ns_child)
            self._propagate_staleness_to_deps(ns_child)

    def _propagate_staleness_to_deps(
        self, dsym: "DataSymbol", skip_seen_check: bool = False
    ) -> None:
        if not skip_seen_check and dsym in self.seen:
            return
        self.seen.add(dsym)
        if (
            dsym not in nbs().updated_symbols
            and dsym not in tracer().this_stmt_updated_symbols
        ):
            if dsym.should_mark_stale(self.updated_sym):
                dsym.fresher_ancestors.add(self.updated_sym)
                dsym.fresher_ancestor_timestamps.add(self.updated_sym.timestamp)
                dsym.required_timestamp = Timestamp.current()
                self._propagate_staleness_to_namespace_parents(
                    dsym, skip_seen_check=True
                )
                self._propagate_staleness_to_namespace_children(
                    dsym, skip_seen_check=True
                )
        for child in self._non_class_to_instance_children(dsym):
            logger.warning("propagate %s %s to %s", dsym, dsym.obj_id, child)
            self._propagate_staleness_to_deps(child)



# -*- coding: utf-8 -*-
import ast
import logging
import re
import shlex
import subprocess
from collections import defaultdict
from contextlib import contextmanager
from typing import (
    cast,
    TYPE_CHECKING,
    Dict,
    FrozenSet,
    Generator,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

from nbsafety.analysis.resolved_symbols import ResolvedDataSymbol
from nbsafety.analysis.live_refs import (
    compute_live_dead_symbol_refs,
    get_symbols_for_references,
    get_live_symbols_and_cells_for_references,
)
from nbsafety.analysis.slicing import CodeCellSlicingMixin
from nbsafety.data_model.timestamp import Timestamp
from nbsafety.ipython_utils import cell_counter as ipy_cell_counter
from nbsafety.run_mode import FlowOrder
from nbsafety.singletons import nbs
from nbsafety.types import CellId, TimestampOrCounter

if TYPE_CHECKING:
    from nbsafety.data_model.data_symbol import DataSymbol


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


_NB_MAGIC_PATTERN = re.compile(r"(^%|^!|^cd |\?$)")


class CheckerResult(NamedTuple):
    live: Set[ResolvedDataSymbol]  # all live symbols in the cell
    used_cells: Set[int]  # last updated timestamps of the live symbols
    live_cells: Set[int]  # cells that define a symbol that was called in the cell
    dead: Set["DataSymbol"]  # symbols that are definitely assigned to
    typechecks: bool  # whether the cell typechecks successfully


def cells() -> Type["ExecutedCodeCell"]:
    return ExecutedCodeCell


class ExecutedCodeCell(CodeCellSlicingMixin):
    _current_cell_by_cell_id: Dict[CellId, "ExecutedCodeCell"] = {}
    _cell_by_cell_ctr: Dict[int, "ExecutedCodeCell"] = {}
    _cell_counter: int = 0
    _position_by_cell_id: Dict[CellId, int] = {}
    _cells_by_tag: Dict[str, Set["ExecutedCodeCell"]] = defaultdict(set)
    _reactive_cells_by_tag: Dict[str, Set[CellId]] = defaultdict(set)

    def __init__(
        self, cell_id: CellId, cell_ctr: int, content: str, tags: Tuple[str, ...]
    ) -> None:
        self.cell_id: CellId = cell_id
        self.cell_ctr: int = cell_ctr
        self.content: str = content
        self.tags: Tuple[str, ...] = tags
        self.reactive_tags: Set[str] = set()
        self._dynamic_parents: Set[CellId] = set()
        self._dynamic_children: Set[CellId] = set()
        self._static_parents: Set[CellId] = set()
        self._static_children: Set[CellId] = set()
        self._used_cell_counters_by_live_symbol: Dict[
            "DataSymbol", Set[int]
        ] = defaultdict(set)
        self._cached_ast: Optional[ast.Module] = None
        self._cached_typecheck_result: Optional[bool] = (
            None if nbs().settings.mark_typecheck_failures_unsafe else True
        )
        self._fresh: bool = False

    @classmethod
    def clear(cls):
        cls._current_cell_by_cell_id = {}
        cls._cell_by_cell_ctr = {}
        cls._cell_counter = 0
        cls._position_by_cell_id = {}
        cls._cells_by_tag.clear()
        cls._reactive_cells_by_tag.clear()

    def __str__(self):
        return self.content

    def __repr__(self):
        return f"<{self.__class__.__name__}[id={self.cell_id},ctr={self.cell_ctr}]>"

    def __hash__(self):
        return hash((self.cell_id, self.cell_ctr))

    def add_used_cell_counter(self, sym: "DataSymbol", ctr: int) -> None:
        if ctr > 0:
            self._used_cell_counters_by_live_symbol[sym].add(ctr)

    def set_fresh(self, new_fresh: bool) -> bool:
        old_fresh = self._fresh
        self._fresh = new_fresh
        return old_fresh

    def mark_as_reactive_for_tag(self, tag: str) -> None:
        self.reactive_tags.add(tag)
        self._reactive_cells_by_tag[tag].add(self.cell_id)

    @classmethod
    def get_reactive_ids_for_tag(cls, tag: str) -> Set[CellId]:
        return cls._reactive_cells_by_tag.get(tag, set())

    def add_dynamic_parent(self, parent: Union["ExecutedCodeCell", CellId]) -> None:
        pid = parent.cell_id if isinstance(parent, ExecutedCodeCell) else parent
        if pid == self.cell_id or pid in self._dynamic_children:
            return
        self._dynamic_parents.add(pid)
        parent = self.from_id(pid)
        parent._dynamic_children.add(self.cell_id)

    def add_static_parent(self, parent: Union["ExecutedCodeCell", CellId]) -> None:
        pid = parent.cell_id if isinstance(parent, ExecutedCodeCell) else parent
        if pid == self.cell_id or pid in self._static_children:
            return
        self._static_parents.add(pid)
        parent = self.from_id(pid)
        parent._static_children.add(self.cell_id)

    @property
    def dynamic_parents(self) -> Generator["ExecutedCodeCell", None, None]:
        for pid in self._dynamic_parents:
            yield self.from_id(pid)

    @property
    def dynamic_children(self) -> Generator["ExecutedCodeCell", None, None]:
        for cid in self._dynamic_children:
            yield self.from_id(cid)

    @property
    def static_parents(self) -> Generator["ExecutedCodeCell", None, None]:
        for pid in self._static_parents:
            yield self.from_id(pid)

    @property
    def static_children(self) -> Generator["ExecutedCodeCell", None, None]:
        for cid in self._static_children:
            yield self.from_id(cid)

    @property
    def dynamic_parent_ids(self) -> FrozenSet[CellId]:
        # trick to catch some mutations at typecheck time w/out runtime overhead
        return cast("FrozenSet[CellId]", self._dynamic_parents)

    @property
    def dynamic_children_ids(self) -> FrozenSet[CellId]:
        return cast("FrozenSet[CellId]", self._dynamic_children)

    @property
    def static_parent_ids(self) -> FrozenSet[CellId]:
        return cast("FrozenSet[CellId]", self._static_parents)

    @property
    def static_children_ids(self) -> FrozenSet[CellId]:
        return cast("FrozenSet[CellId]", self._static_children)

    @classmethod
    def create_and_track(
        cls,
        cell_id: CellId,
        content: str,
        tags: Tuple[str, ...],
        validate_ipython_counter: bool = True,
    ) -> "ExecutedCodeCell":
        cls._cell_counter += 1
        cell_ctr = cls._cell_counter
        if validate_ipython_counter:
            assert cell_ctr == ipy_cell_counter()
        prev_cell = cls.from_id(cell_id)
        cell = cls(cell_id, cell_ctr, content, tags)
        if prev_cell is not None:
            cell._dynamic_children = prev_cell._dynamic_children
            cell._static_children = prev_cell._static_children
            for tag in prev_cell.tags:
                cls._cells_by_tag[tag].discard(prev_cell)
            for tag in prev_cell.reactive_tags:
                cls._reactive_cells_by_tag[tag].discard(prev_cell.cell_id)
        for tag in tags:
            cls._cells_by_tag[tag].add(cell)
        cls._cell_by_cell_ctr[cell_ctr] = cell
        cur_cell = cls._current_cell_by_cell_id.get(cell_id, None)
        cur_cell_ctr = None if cur_cell is None else cur_cell.cell_ctr
        if cur_cell_ctr is None or cell_ctr > cur_cell_ctr:
            cls._current_cell_by_cell_id[cell_id] = cell
        return cell

    @classmethod
    def set_cell_positions(cls, order_index_by_cell_id: Dict[CellId, int]):
        cls._position_by_cell_id = order_index_by_cell_id

    @classmethod
    @contextmanager
    def _override_position_index_for_current_flow_semantics(
        cls,
    ) -> Generator[None, None, None]:
        orig_position_by_cell_id = cls._position_by_cell_id
        try:
            if nbs().mut_settings.flow_order == FlowOrder.ANY_ORDER:
                cls.set_cell_positions({})
            yield
        finally:
            cls.set_cell_positions(orig_position_by_cell_id)

    @property
    def position(self) -> int:
        return self._position_by_cell_id.get(self.cell_id, -1)

    @classmethod
    def exec_counter(cls) -> int:
        return cls._cell_counter

    @classmethod
    def next_exec_counter(cls) -> int:
        return cls.exec_counter() + 1

    @classmethod
    def all_cells_most_recently_run_for_each_id(
        cls,
    ) -> Generator["ExecutedCodeCell", None, None]:
        yield from cls._current_cell_by_cell_id.values()

    @classmethod
    def from_timestamp(cls, ts: TimestampOrCounter) -> "ExecutedCodeCell":
        if isinstance(ts, Timestamp):
            return cls._cell_by_cell_ctr[ts.cell_num]
        else:
            return cls._cell_by_cell_ctr[ts]

    @classmethod
    def from_id(cls, cell_id: CellId) -> Optional["ExecutedCodeCell"]:
        return cls._current_cell_by_cell_id.get(cell_id, None)

    @classmethod
    def from_tag(cls, tag: str) -> Set["ExecutedCodeCell"]:
        return cls._cells_by_tag.get(tag, set())

    def sanitized_content(self):
        lines = []
        for line in self.content.strip().split("\n"):
            # TODO: figure out more robust strategy for filtering / transforming lines for the ast parser
            # we filter line magics, but for %time, we would ideally like to trace the statement being timed
            # TODO: how to do this?
            if _NB_MAGIC_PATTERN.search(line.strip()) is None:
                lines.append(line)
        return "\n".join(lines)

    def to_ast(self, override: Optional[ast.Module] = None) -> ast.Module:
        if override is not None:
            self._cached_ast = override
        if self._cached_ast is None:
            self._cached_ast = ast.parse(self.sanitized_content())
        return self._cached_ast

    @property
    def is_current_for_id(self) -> bool:
        return self._current_cell_by_cell_id.get(self.cell_id, None) is self

    @classmethod
    def current_cell(cls) -> "ExecutedCodeCell":
        return cls._cell_by_cell_ctr[cls._cell_counter]

    def get_max_used_live_symbol_cell_counter(
        self, live_symbols: Set[ResolvedDataSymbol], filter_to_reactive: bool = False
    ) -> int:
        with self._override_position_index_for_current_flow_semantics():
            max_used_cell_ctr = -1
            this_cell_pos = self.position
            for sym in live_symbols:
                if filter_to_reactive:
                    if sym.is_blocking:
                        continue
                    if (
                        not sym.is_reactive
                        and sym.dsym not in nbs().updated_reactive_symbols
                    ):
                        continue
                for cell_ctr in self._used_cell_counters_by_live_symbol.get(
                    sym.dsym, []
                ):
                    if self.from_timestamp(cell_ctr).position <= this_cell_pos:
                        max_used_cell_ctr = max(max_used_cell_ctr, cell_ctr)
            return max_used_cell_ctr

    def check_and_resolve_symbols(
        self, update_liveness_time_versions: bool = False
    ) -> CheckerResult:
        live_symbol_refs, dead_symbol_refs = compute_live_dead_symbol_refs(
            self.to_ast(), scope=nbs().global_scope
        )
        live_resolved_symbols, live_cells = get_live_symbols_and_cells_for_references(
            live_symbol_refs,
            nbs().global_scope,
            self.cell_ctr,
            update_liveness_time_versions=update_liveness_time_versions,
        )
        # only mark dead attrsubs as killed if we can traverse the entire chain
        dead_symbols, _ = get_symbols_for_references(
            dead_symbol_refs, nbs().global_scope
        )
        for resolved in live_resolved_symbols:
            if resolved.is_deep:
                resolved.dsym.cells_where_deep_live.add(self)
                self.add_used_cell_counter(
                    resolved.dsym, resolved.dsym.timestamp.cell_num
                )
            else:
                resolved.dsym.cells_where_shallow_live.add(self)
                self.add_used_cell_counter(
                    resolved.dsym,
                    resolved.dsym.timestamp_excluding_ns_descendents.cell_num,
                )
        return CheckerResult(
            live=live_resolved_symbols,
            used_cells={
                resolved.dsym.timestamp.cell_num
                if resolved.is_deep
                else resolved.dsym.timestamp_excluding_ns_descendents.cell_num
                for resolved in live_resolved_symbols
            },
            live_cells=live_cells,
            dead=dead_symbols,
            typechecks=self._typechecks(live_cells, live_resolved_symbols),
        )

    def compute_phantom_cell_info(self, used_cells: Set[int]) -> Dict[CellId, Set[int]]:
        used_cell_counters_by_cell_id = defaultdict(set)
        used_cell_counters_by_cell_id[self.cell_id].add(self.exec_counter())
        for cell_num in used_cells:
            used_cell_counters_by_cell_id[self.from_timestamp(cell_num).cell_id].add(
                cell_num
            )
        return {
            cell_id: cell_execs
            for cell_id, cell_execs in used_cell_counters_by_cell_id.items()
            if len(cell_execs) >= 2
        }

    def _build_typecheck_slice(
        self, live_cell_ctrs: Set[int], live_symbols: Set[ResolvedDataSymbol]
    ) -> str:
        # TODO: typecheck statically-resolvable nested symbols too, not just top-level
        live_cell_counters = {self.cell_ctr}
        for live_cell_num in live_cell_ctrs:
            if self.from_timestamp(live_cell_num).is_current_for_id:
                live_cell_counters.add(live_cell_num)
        live_cells = [self.from_timestamp(ctr) for ctr in sorted(live_cell_counters)]
        top_level_symbols = {sym.dsym.get_top_level() for sym in live_symbols}
        top_level_symbols.discard(None)
        return "{type_declarations}\n\n{content}".format(
            type_declarations="\n".join(
                f"{sym.name}: {sym.get_type_annotation_string()}"
                for sym in top_level_symbols
            ),
            content="\n".join(
                live_cell.sanitized_content() for live_cell in live_cells
            ),
        )

    def _typechecks(
        self, live_cell_ctrs: Set[int], live_symbols: Set[ResolvedDataSymbol]
    ) -> bool:
        if self._cached_typecheck_result is not None:
            return self._cached_typecheck_result
        typecheck_slice = self._build_typecheck_slice(live_cell_ctrs, live_symbols)
        try:
            # TODO: parse the output in order to pass up to the user
            ret = subprocess.call(f"mypy -c {shlex.quote(typecheck_slice)}", shell=True)
            self._cached_typecheck_result = ret == 0
        except Exception:
            logger.exception("Exception occurred during type checking")
            self._cached_typecheck_result = True
        return self._cached_typecheck_result

    @property
    def needs_typecheck(self):
        return self._cached_typecheck_result is None

    def invalidate_typecheck_result(self):
        self._cached_typecheck_result = None



from typing import Tuple, TypeVar, Union
from nbsafety.data_model.timestamp import Timestamp

CellId = Union[str, int]
SupportedIndexType = Union[str, int, Tuple[Union[str, int], ...]]
TimestampOrCounter = TypeVar("TimestampOrCounter", Timestamp, int)



# -*- coding: utf-8 -*-
import ast
import logging
import traceback
from typing import cast

from nbsafety.singletons import nbs
from nbsafety.data_model.code_cell import cells
from pyccolo import AstRewriter


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class SafetyAstRewriter(AstRewriter):
    def visit(self, node: ast.AST):
        try:
            ret = super().visit(node)
            cells().current_cell().to_ast(
                override=cast(ast.Module, self.orig_to_copy_mapping[id(node)])
            )
            return ret
        except Exception as e:
            nbs().set_exception_raised_during_execution(e)
            traceback.print_exc()
            raise e



# -*- coding: utf-8 -*-
import json
import os
import platform
import sys
import argparse

from jupyter_client.kernelspec import KernelSpecManager
from IPython.utils.tempdir import TemporaryDirectory

PACKAGE = __package__.split(".")[0]
DISPLAY_NAME = f"Python 3 ({PACKAGE})"
kernel_json = {
    "argv": [
        sys.executable,
        "-m",
        "nbsafety.kernel",
        "-f",
        "{connection_file}",
    ],
    "display_name": DISPLAY_NAME,
    "language": "python",
    "codemirror_mode": "shell",
}


def install_my_kernel_spec(user=True, prefix=None):
    with TemporaryDirectory() as td:
        os.chmod(td, 0o755)  # Starts off as 700, not user readable
        cp = "cp"
        if platform.system().lower().startswith("win"):
            cp = "copy"
        import nbsafety

        resources = os.path.join(
            os.path.dirname(os.path.abspath(nbsafety.__file__)), "resources", "kernel"
        )
        logo32 = os.path.join(resources, "logo-32x32.png")
        logo64 = os.path.join(resources, "logo-64x64.png")
        os.system(f"{cp} {logo32} {logo64} {td}")
        with open(os.path.join(td, "kernel.json"), "w") as f:
            json.dump(kernel_json, f, sort_keys=True)

        print(f"Installing KernelSpec for {PACKAGE} kernel")
        KernelSpecManager().install_kernel_spec(
            td, kernel_name=f"{PACKAGE}", user=user, prefix=prefix
        )


def _is_root():
    try:
        return os.geteuid() == 0
    except AttributeError:
        return False  # assume not an admin on non-Unix platforms


def main(argv=None):
    parser = argparse.ArgumentParser(
        description=f"Install KernelSpec for {PACKAGE} kernel."
    )
    prefix_locations = parser.add_mutually_exclusive_group()

    prefix_locations.add_argument(
        "--user", help="Install KernelSpec in user homedirectory", action="store_true"
    )
    prefix_locations.add_argument(
        "--sys-prefix",
        help="Install KernelSpec in sys.prefix. Useful in conda / virtualenv",
        action="store_true",
        dest="sys_prefix",
    )
    prefix_locations.add_argument(
        "--prefix", help="Install KernelSpec in this prefix", default=None
    )

    args = parser.parse_args(argv)

    user = False
    prefix = None
    if args.sys_prefix:
        prefix = sys.prefix
    elif args.prefix:
        prefix = args.prefix
    elif args.user or not _is_root():
        user = True

    install_my_kernel_spec(user=user, prefix=prefix)
    return 0


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
import logging
from typing import cast, TYPE_CHECKING, Optional
from nbsafety.run_mode import ExecutionMode
from nbsafety.singletons import nbs
from nbsafety.tracing.mutation_event import resolve_mutating_method
from nbsafety.utils import CommonEqualityMixin

if TYPE_CHECKING:
    from nbsafety.analysis.symbol_ref import Atom
    from nbsafety.data_model.data_symbol import DataSymbol


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class ResolvedDataSymbol(CommonEqualityMixin):
    def __init__(
        self,
        dsym: "DataSymbol",
        atom: "Atom",
        next_atom: Optional["Atom"],
        liveness_timestamp: Optional[int] = None,
    ) -> None:
        self.dsym = dsym
        self.atom = atom
        self.next_atom = next_atom
        self.liveness_timestamp = liveness_timestamp

    def __hash__(self):
        return hash(
            (
                self.dsym,
                self.atom,
                self.next_atom,
                self.liveness_timestamp,
            )
        )

    @property
    def timestamp(self):
        return self.dsym.timestamp

    @property
    def is_called(self):
        return self.atom.is_callpoint

    @property
    def is_last(self):
        return self.next_atom is None

    @property
    def is_reactive(self):
        if self.is_blocking:
            return False
        return self.atom.is_reactive or (
            self.is_live and self.dsym in nbs().updated_deep_reactive_symbols
        )

    @property
    def is_blocking(self):
        return self.atom.is_blocking or (
            self.is_live
            and nbs().blocked_reactive_timestamps_by_symbol.get(self.dsym, -1)
            >= self.dsym.timestamp.cell_num
        )

    @property
    def is_dead(self):
        return self.liveness_timestamp is None

    @property
    def is_live(self):
        return not self.is_dead

    @property
    def is_deep(self) -> bool:
        # for live symbols, if it is used in its entirety
        assert self.is_live
        if self.is_reactive:
            return True
        if self.next_atom is None:
            return True
        elif not self.next_atom.is_callpoint:
            return False
        elif self.is_mutating:  # self.next_atom.is_callpoint
            return False
        else:
            return True

    @property
    def is_shallow(self) -> bool:
        # for live symbols, if only a portion (attr or subscript) is used
        assert self.is_live
        return not self.is_deep

    @property
    def is_mutating(self) -> bool:
        assert self.is_live
        if not self.next_atom.is_callpoint:
            return False
        return (
            resolve_mutating_method(self.dsym.obj, cast(str, self.next_atom.value))
            is not None
        )

    @property
    def is_unsafe(self) -> bool:
        assert self.is_live
        if self.next_atom is None:
            return False
        if self.next_atom.is_callpoint:
            if self.is_mutating and (
                self.is_reactive
                or nbs().mut_settings.exec_mode == ExecutionMode.REACTIVE
            ):
                return True
            else:
                return False
        if (
            isinstance(self.dsym.obj, (list, tuple))
            and isinstance(self.next_atom.value, int)
            and self.next_atom.value >= len(self.dsym.obj)
        ):
            return True
        if (
            isinstance(self.dsym.obj, dict)
            and self.next_atom.value not in self.dsym.obj
        ):
            return True
        if (
            not isinstance(self.dsym.obj, (dict, list, tuple))
            and isinstance(self.next_atom.value, str)
            and (
                # the first check guards against properties; hasattr actually executes code for those if called
                # on the actual object, which we want to avoid
                not hasattr(self.dsym.obj.__class__, self.next_atom.value)
                and not hasattr(self.dsym.obj, self.next_atom.value)
            )
        ):
            # TODO: fix this once we can distinguish between attrs and subscripts in the chain
            return True
        return False

    def is_stale_at_position(self, pos: int):
        return self.dsym.is_stale_at_position(pos, deep=self.is_deep)



# -*- coding: utf-8 -*-
import ast


def subscript_to_slice(node: ast.Subscript) -> ast.expr:
    if isinstance(node.slice, ast.Index):
        return node.slice.value  # type: ignore
    else:
        return node.slice  # type: ignore



# -*- coding: utf-8 -*-
import ast
import textwrap
from typing import Dict

from nbsafety.singletons import tracer
from pyccolo.stmt_mapper import StatementMapper
from .utils import make_safety_fixture


_safety_fixture, _ = make_safety_fixture()


def compute_lineno_to_stmt_mapping(code: str) -> Dict[int, ast.stmt]:
    mapper = StatementMapper({}, [tracer()], {})
    mapper(ast.parse(textwrap.dedent(code).strip()))
    return mapper.line_to_stmt_map


def test_for_loop():
    code = """
        for i in range(10):
            a: int = i
            b = a + i
            lst: List[int] = [a, b]
        """
    mapping = compute_lineno_to_stmt_mapping(code)
    assert isinstance(mapping[1], ast.For)
    assert isinstance(mapping[2], ast.AnnAssign)
    assert isinstance(mapping[3], ast.Assign)
    assert isinstance(mapping[4], ast.AnnAssign)


def test_multiline_for_loop():
    code = """
        for i in [
            0,
            1,
            2,
            3,
            4,
        ]:
            a = i
            b = a + i
            lst = [a, b]
        """
    mapping = compute_lineno_to_stmt_mapping(code)
    # for i in range(1, 7):
    #     assert isinstance(mapping[i], ast.For)
    # assert 7 not in mapping
    assert isinstance(mapping[1], ast.For)
    for i in range(2, 8):
        assert i not in mapping
    assert isinstance(mapping[8], ast.Assign)
    assert isinstance(mapping[9], ast.Assign)
    assert isinstance(mapping[10], ast.Assign)


def test_if():
    code = """
        if True:
            x = 0
        else:
            x: int = 0
        """
    mapping = compute_lineno_to_stmt_mapping(code)
    assert isinstance(mapping[1], ast.If)
    assert isinstance(mapping[2], ast.Assign)
    assert 3 not in mapping
    assert isinstance(mapping[4], ast.AnnAssign)



# -*- coding: utf-8 -*-
import logging
from typing import Set

from nbsafety.data_model.code_cell import cells
from nbsafety.singletons import nbs
from nbsafety.types import CellId
from test.utils import make_safety_fixture, skipif_known_failing

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture(mark_typecheck_failures_unsafe=True)


def run_cell(cell, cell_id=None, **kwargs):
    """Mocks the `change active cell` portion of the comm protocol"""
    if cell_id is not None:
        nbs().handle({"type": "change_active_cell", "active_cell_id": cell_id})
    run_cell_(cell, **kwargs)


def get_cell_ids_needing_typecheck() -> Set[CellId]:
    return {
        cell.cell_id
        for cell in cells().all_cells_most_recently_run_for_each_id()
        if cell.needs_typecheck
    }


def test_int_change_to_str_triggers_typecheck():
    run_cell("a = 1", 1)
    assert not get_cell_ids_needing_typecheck()
    run_cell("b = 2", 2)
    assert not get_cell_ids_needing_typecheck()
    run_cell("logging.info(a + b)", 3)
    assert not get_cell_ids_needing_typecheck()
    run_cell('b = "b"', 4)
    assert get_cell_ids_needing_typecheck() == {3}
    nbs().check_and_link_multiple_cells()
    assert not get_cell_ids_needing_typecheck()
    assert cells().from_id(3)._cached_typecheck_result is False



# -*- coding: utf-8 -*-
import logging

from nbsafety.singletons import nbs
from test.utils import make_safety_fixture

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture()


def updated_symbol_names():
    return sorted(
        [sym.readable_name for sym in nbs().updated_symbols if not sym.is_anonymous]
    )


def run_cell(cell):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell)


def test_simplest():
    run_cell("a = 0")
    assert updated_symbol_names() == ["a"]
    run_cell("a += 1")
    assert updated_symbol_names() == ["a"]


def test_dict_hierarchy():
    run_cell("d = {}")
    updated_sym_names = updated_symbol_names()
    assert updated_sym_names == ["d"], "got %s" % updated_sym_names
    run_cell('d["foo"] = {}')
    assert updated_symbol_names() == sorted(["d[foo]", "d"])
    run_cell('d["foo"]["bar"] = []')
    updated_sym_names = updated_symbol_names()
    assert updated_sym_names == sorted(["d[foo][bar]", "d[foo]", "d"]), (
        "got %s" % updated_sym_names
    )
    run_cell('d["foo"]["bar"] = 0')
    assert updated_symbol_names() == sorted(["d[foo][bar]", "d[foo]", "d"])



# -*- coding: utf-8 -*-
import asyncio
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
import inspect
import json
import logging
import re
from types import FrameType
from typing import (
    cast,
    Any,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Set,
    Optional,
    Tuple,
    Union,
)

from IPython import get_ipython
from IPython.core.magic import register_cell_magic, register_line_magic

import pyccolo as pyc

from nbsafety.ipython_utils import (
    ast_transformer_context,
    input_transformer_context,
    run_cell,
    save_number_of_currently_executing_cell,
)
from nbsafety import line_magics
from nbsafety.data_model.code_cell import cells, ExecutedCodeCell
from nbsafety.data_model.data_symbol import DataSymbol
from nbsafety.data_model.namespace import Namespace
from nbsafety.data_model.scope import Scope
from nbsafety.data_model.timestamp import Timestamp
from nbsafety.run_mode import ExecutionMode, ExecutionSchedule, FlowOrder, SafetyRunMode
from nbsafety import singletons
from nbsafety.tracing.nbsafety_tracer import SafetyTracer
from nbsafety.types import CellId, SupportedIndexType

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

_SAFETY_LINE_MAGIC = "safety"
_NB_MAGIC_PATTERN = re.compile(r"(^%|^!|^cd |\?$)")


class NotebookSafetySettings(NamedTuple):
    store_history: bool
    test_context: bool
    use_comm: bool
    track_dependencies: bool
    mark_stale_symbol_usages_unsafe: bool
    mark_typecheck_failures_unsafe: bool
    mark_phantom_cell_usages_unsafe: bool
    enable_reactive_modifiers: bool
    mode: SafetyRunMode


@dataclass
class MutableNotebookSafetySettings:
    trace_messages_enabled: bool
    highlights_enabled: bool
    static_slicing_enabled: bool
    dynamic_slicing_enabled: bool
    exec_mode: ExecutionMode
    exec_schedule: ExecutionSchedule
    flow_order: FlowOrder


class FrontendCheckerResult(NamedTuple):
    stale_cells: Set[CellId]
    fresh_cells: Set[CellId]
    new_fresh_cells: Set[CellId]
    forced_reactive_cells: Set[CellId]
    stale_links: Dict[CellId, Set[CellId]]
    refresher_links: Dict[CellId, Set[CellId]]
    phantom_cell_info: Dict[CellId, Dict[CellId, Set[int]]]

    def to_json(self) -> Dict[str, Any]:
        return {
            "stale_cells": list(self.stale_cells),
            "fresh_cells": list(self.fresh_cells),
            "new_fresh_cells": list(self.new_fresh_cells),
            "forced_reactive_cells": list(self.forced_reactive_cells),
            "stale_links": {
                cell_id: list(linked_cell_ids)
                for cell_id, linked_cell_ids in self.stale_links.items()
            },
            "refresher_links": {
                cell_id: list(linked_cell_ids)
                for cell_id, linked_cell_ids in self.refresher_links.items()
            },
        }


class NotebookSafety(singletons.NotebookSafety):
    """Holds all the state necessary to detect stale dependencies in Jupyter notebooks."""

    def __init__(self, cell_magic_name=None, use_comm=False, settrace=None, **kwargs):
        super().__init__()
        cells().clear()
        self.settings: NotebookSafetySettings = NotebookSafetySettings(
            store_history=kwargs.pop("store_history", True),
            test_context=kwargs.pop("test_context", False),
            use_comm=use_comm,
            track_dependencies=True,
            mark_stale_symbol_usages_unsafe=kwargs.pop(
                "mark_stale_symbol_usages_unsafe", True
            ),
            mark_typecheck_failures_unsafe=kwargs.pop(
                "mark_typecheck_failures_unsafe", False
            ),
            mark_phantom_cell_usages_unsafe=kwargs.pop(
                "mark_phantom_cell_usages_unsafe", False
            ),
            enable_reactive_modifiers=kwargs.pop("enable_reactive_modifiers", True),
            mode=SafetyRunMode.get(),
        )
        self.mut_settings: MutableNotebookSafetySettings = (
            MutableNotebookSafetySettings(
                trace_messages_enabled=kwargs.pop("trace_messages_enabled", False),
                highlights_enabled=kwargs.pop("highlights_enabled", True),
                static_slicing_enabled=kwargs.pop("static_slicing_enabled", True),
                dynamic_slicing_enabled=kwargs.pop("dynamic_slicing_enabled", True),
                exec_mode=ExecutionMode(kwargs.pop("exec_mode", ExecutionMode.NORMAL)),
                exec_schedule=ExecutionSchedule(
                    kwargs.pop("exec_schedule", ExecutionSchedule.LIVENESS_BASED)
                ),
                flow_order=FlowOrder(kwargs.pop("flow_order", FlowOrder.ANY_ORDER)),
            )
        )
        # Note: explicitly adding the types helps PyCharm intellisense
        self.namespaces: Dict[int, Namespace] = {}
        # TODO: wrap this in something that clears the dict entry when the set is 0 length
        self.aliases: Dict[int, Set[DataSymbol]] = defaultdict(set)
        self.dynamic_data_deps: Dict[Timestamp, Set[Timestamp]] = defaultdict(set)
        self.static_data_deps: Dict[Timestamp, Set[Timestamp]] = defaultdict(set)
        self.global_scope: Scope = Scope()
        self.updated_symbols: Set[DataSymbol] = set()
        self.updated_reactive_symbols: Set[DataSymbol] = set()
        self.updated_deep_reactive_symbols: Set[DataSymbol] = set()
        self.blocked_reactive_timestamps_by_symbol: Dict[DataSymbol, int] = {}
        self.statement_to_func_cell: Dict[int, DataSymbol] = {}
        self._active_cell_id: Optional[CellId] = None
        self.safety_issue_detected = False
        if cell_magic_name is None:
            self._cell_magic = None
        else:
            self._cell_magic = self._make_cell_magic(cell_magic_name)
        self._line_magic = self._make_line_magic()
        self._prev_cell_stale_symbols: Set[DataSymbol] = set()
        self._cell_name_to_cell_num_mapping: Dict[str, int] = {}
        self._exception_raised_during_execution: Optional[Exception] = None
        self._saved_debug_message: Optional[str] = None
        self.min_timestamp = -1
        self._tags: Tuple[str, ...] = ()
        if use_comm:
            get_ipython().kernel.comm_manager.register_target(
                __package__, self._comm_target
            )

    @property
    def is_develop(self) -> bool:
        return self.settings.mode == SafetyRunMode.DEVELOP

    @property
    def is_test(self) -> bool:
        return self.settings.test_context

    @property
    def trace_messages_enabled(self) -> bool:
        return self.mut_settings.trace_messages_enabled

    @trace_messages_enabled.setter
    def trace_messages_enabled(self, new_val) -> None:
        self.mut_settings.trace_messages_enabled = new_val

    def get_first_full_symbol(self, obj_id: int) -> Optional[DataSymbol]:
        # TODO: also avoid anonymous namespaces?
        for alias in self.aliases.get(obj_id, []):
            if not alias.is_anonymous:
                return alias
        return None

    @staticmethod
    def cell_counter() -> int:
        return cells().exec_counter()

    def add_dynamic_data_dep(self, child: Timestamp, parent: Timestamp):
        self.dynamic_data_deps[child].add(parent)
        cells().from_timestamp(child).add_dynamic_parent(cells().from_timestamp(parent))

    def add_static_data_dep(self, child: Timestamp, parent: Timestamp):
        self.static_data_deps[child].add(parent)
        cells().from_timestamp(child).add_static_parent(cells().from_timestamp(parent))

    def reset_cell_counter(self):
        # only called in test context
        assert not self.settings.store_history
        self.dynamic_data_deps.clear()
        self.static_data_deps.clear()
        for sym in self.all_data_symbols():
            sym._timestamp = (
                sym._max_inner_timestamp
            ) = sym.required_timestamp = Timestamp.uninitialized()
            sym.timestamp_by_used_time.clear()
            sym.timestamp_by_liveness_time.clear()
        cells().clear()

    def set_exception_raised_during_execution(
        self, new_val: Optional[Exception] = None
    ) -> Optional[Exception]:
        ret = self._exception_raised_during_execution
        self._exception_raised_during_execution = new_val
        return ret

    def get_position(self, frame: FrameType) -> Tuple[Optional[int], int]:
        try:
            cell_num = self._cell_name_to_cell_num_mapping.get(
                frame.f_code.co_filename, None
            )
            if cell_num is None:
                cell_num = self.cell_counter()
            return cell_num, frame.f_lineno
        except KeyError as e:
            logger.error(
                "key error while retrieving cell for %s", frame.f_code.co_filename
            )
            raise e

    def set_name_to_cell_num_mapping(self, frame: FrameType):
        self._cell_name_to_cell_num_mapping[
            frame.f_code.co_filename
        ] = cells().exec_counter()

    def is_cell_file(self, fname: str) -> bool:
        return fname in self._cell_name_to_cell_num_mapping

    def set_active_cell(self, cell_id: CellId) -> None:
        self._active_cell_id = cell_id

    def set_tags(self, tags: Tuple[str, ...]) -> None:
        self._tags = tags

    def reactivity_cleanup(self) -> None:
        for cell in cells().all_cells_most_recently_run_for_each_id():
            cell.set_fresh(False)

    def _comm_target(self, comm, open_msg) -> None:
        @comm.on_msg
        def _responder(msg):
            request = msg["content"]["data"]
            self.handle(request, comm=comm)

        comm.send({"type": "establish"})

    def handle(self, request, comm=None) -> None:
        if request["type"] == "change_active_cell":
            self.set_active_cell(request["active_cell_id"])
        elif request["type"] == "cell_freshness":
            if self._active_cell_id is None:
                self.set_active_cell(request.get("executed_cell_id", None))
            cell_id = request.get("executed_cell_id", None)
            order_index_by_id = request["order_index_by_cell_id"]
            cells().set_cell_positions(order_index_by_id)
            cells_to_check = (
                cell
                for cell in (cells().from_id(cell_id) for cell_id in order_index_by_id)
                if cell is not None
            )
            response = self.check_and_link_multiple_cells(
                cells_to_check=cells_to_check, last_executed_cell_id=cell_id
            ).to_json()
            response["type"] = "cell_freshness"
            response["exec_mode"] = self.mut_settings.exec_mode.value
            response["exec_schedule"] = self.mut_settings.exec_schedule.value
            response["flow_order"] = self.mut_settings.flow_order.value
            response["last_executed_cell_id"] = cell_id
            response["highlights_enabled"] = self.mut_settings.highlights_enabled
            if comm is not None:
                comm.send(response)
        elif request["type"] == "reactivity_cleanup":
            self.reactivity_cleanup()
        else:
            dbg_msg = "Unsupported request type for request %s" % request
            logger.error(dbg_msg)
            self._saved_debug_message = dbg_msg

    def check_and_link_multiple_cells(
        self,
        cells_to_check: Optional[Iterable[ExecutedCodeCell]] = None,
        update_liveness_time_versions: bool = False,
        last_executed_cell_id: Optional[CellId] = None,
    ) -> FrontendCheckerResult:
        SafetyTracer.instance()  # force initialization here in case not already inited
        stale_cells = set()
        unsafe_order_cells: Set[CellId] = set()
        typecheck_error_cells = set()
        fresh_cells = set()
        new_fresh_cells = set()
        forced_reactive_cells = set()
        stale_symbols_by_cell_id: Dict[CellId, Set[DataSymbol]] = {}
        killing_cell_ids_for_symbol: Dict[DataSymbol, Set[CellId]] = defaultdict(set)
        phantom_cell_info: Dict[CellId, Dict[CellId, Set[int]]] = {}
        checker_results_by_cid = {}
        if last_executed_cell_id is None:
            last_executed_cell = None
            last_executed_cell_pos = None
        else:
            last_executed_cell = cells().from_id(last_executed_cell_id)
            last_executed_cell_pos = last_executed_cell.position
            for tag in last_executed_cell.tags:
                for reactive_cell_id in cells().get_reactive_ids_for_tag(tag):
                    forced_reactive_cells.add(reactive_cell_id)
        if cells_to_check is None:
            cells_to_check = cells().all_cells_most_recently_run_for_each_id()
        cells_to_check = sorted(cells_to_check, key=lambda c: c.position)
        for cell in cells_to_check:
            try:
                checker_result = cell.check_and_resolve_symbols(
                    update_liveness_time_versions=update_liveness_time_versions
                )
            except SyntaxError:
                continue
            cell_id = cell.cell_id
            checker_results_by_cid[cell_id] = checker_result
            # if self.mut_settings.flow_order == FlowOrder.IN_ORDER:
            #     for live_sym in checker_result.live:
            #         if cells().from_timestamp(live_sym.timestamp).position > cell.position:
            #             unsafe_order_cells.add(cell_id)
            #             break
            if self.mut_settings.flow_order == FlowOrder.IN_ORDER:
                if (
                    last_executed_cell_pos is not None
                    and cell.position <= last_executed_cell_pos
                ):
                    continue
            if self.mut_settings.exec_schedule == ExecutionSchedule.LIVENESS_BASED:
                stale_symbols = {
                    sym.dsym
                    for sym in checker_result.live
                    if sym.is_stale_at_position(cell.position)
                }
            else:
                stale_symbols = set()
            if len(stale_symbols) > 0:
                stale_symbols_by_cell_id[cell_id] = stale_symbols
                stale_cells.add(cell_id)
            if not checker_result.typechecks:
                typecheck_error_cells.add(cell_id)
            for dead_sym in checker_result.dead:
                killing_cell_ids_for_symbol[dead_sym].add(cell_id)

            is_fresh = cell_id not in stale_cells
            if self.settings.mark_phantom_cell_usages_unsafe:
                phantom_cell_info_for_cell = cell.compute_phantom_cell_info(
                    checker_result.used_cells
                )
                if len(phantom_cell_info_for_cell) > 0:
                    phantom_cell_info[cell_id] = phantom_cell_info_for_cell
            if self.mut_settings.exec_schedule == ExecutionSchedule.DAG_BASED:
                is_fresh = False
                flow_order = self.mut_settings.flow_order
                if self.mut_settings.dynamic_slicing_enabled:
                    for par in cell.dynamic_parents:
                        if (
                            flow_order == flow_order.IN_ORDER
                            and par.position >= cell.position
                        ):
                            continue
                        if par.cell_ctr > max(cell.cell_ctr, self.min_timestamp):
                            is_fresh = True
                            break
                if not is_fresh and self.mut_settings.static_slicing_enabled:
                    for par in cell.static_parents:
                        if (
                            flow_order == flow_order.IN_ORDER
                            and par.position >= cell.position
                        ):
                            continue
                        if par.cell_ctr > max(cell.cell_ctr, self.min_timestamp):
                            is_fresh = True
                            break
            else:
                is_fresh = is_fresh and (
                    cell.get_max_used_live_symbol_cell_counter(checker_result.live)
                    > max(cell.cell_ctr, self.min_timestamp)
                )
            if self.mut_settings.exec_schedule == ExecutionSchedule.STRICT:
                for dead_sym in checker_result.dead:
                    if dead_sym.timestamp.cell_num > max(
                        cell.cell_ctr, self.min_timestamp
                    ):
                        is_fresh = True
            if is_fresh:
                fresh_cells.add(cell_id)
            if not cells().from_id(cell_id).set_fresh(is_fresh) and is_fresh:
                new_fresh_cells.add(cell_id)
            if is_fresh and self.mut_settings.exec_schedule == ExecutionSchedule.STRICT:
                break
        if self.mut_settings.exec_schedule == ExecutionSchedule.DAG_BASED:
            prev_stale_cells: Set[CellId] = set()
            while True:
                for cell in cells_to_check:
                    if cell.cell_id in stale_cells:
                        continue
                    if self.mut_settings.dynamic_slicing_enabled:
                        if cell.dynamic_parent_ids & (fresh_cells | stale_cells):
                            stale_cells.add(cell.cell_id)
                            continue
                    if self.mut_settings.static_slicing_enabled:
                        if cell.static_parent_ids & (fresh_cells | stale_cells):
                            stale_cells.add(cell.cell_id)
                if prev_stale_cells == stale_cells:
                    break
                prev_stale_cells = set(stale_cells)
            fresh_cells -= stale_cells
            new_fresh_cells -= stale_cells
            for cell_id in stale_cells:
                cells().from_id(cell_id).set_fresh(False)
        if self.mut_settings.exec_mode != ExecutionMode.REACTIVE:
            for cell_id in new_fresh_cells:
                if cell_id not in checker_results_by_cid:
                    continue
                cell = cells().from_id(cell_id)
                if cell.get_max_used_live_symbol_cell_counter(
                    checker_results_by_cid[cell_id].live, filter_to_reactive=True
                ) > max(cell.cell_ctr, self.min_timestamp):
                    forced_reactive_cells.add(cell_id)
        stale_links: Dict[CellId, Set[CellId]] = defaultdict(set)
        refresher_links: Dict[CellId, Set[CellId]] = defaultdict(set)
        eligible_refresher_for_dag = fresh_cells | stale_cells
        for stale_cell_id in stale_cells:
            refresher_cell_ids: Set[CellId] = set()
            if self.mut_settings.flow_order == ExecutionSchedule.DAG_BASED:
                if self.mut_settings.dynamic_slicing_enabled:
                    refresher_cell_ids |= (
                        cells().from_id(stale_cell_id).dynamic_parent_ids
                        & eligible_refresher_for_dag
                    )
                if self.mut_settings.static_slicing_enabled:
                    refresher_cell_ids |= (
                        cells().from_id(stale_cell_id).static_parent_ids
                        & eligible_refresher_for_dag
                    )
            else:
                stale_syms = stale_symbols_by_cell_id.get(stale_cell_id, set())
                refresher_cell_ids = refresher_cell_ids.union(
                    *(
                        killing_cell_ids_for_symbol[stale_sym]
                        for stale_sym in stale_syms
                    )
                )
            if self.mut_settings.flow_order == FlowOrder.IN_ORDER:
                refresher_cell_ids = {
                    cid
                    for cid in refresher_cell_ids
                    if cells().from_id(cid).position
                    < cells().from_id(stale_cell_id).position
                }
            if last_executed_cell_id is not None:
                refresher_cell_ids.discard(last_executed_cell_id)
            stale_links[stale_cell_id] = refresher_cell_ids
        stale_link_changes = True
        # transitive closer up until we hit non-stale refresher cells
        while stale_link_changes:
            stale_link_changes = False
            for stale_cell_id in stale_cells:
                new_stale_links = set(stale_links[stale_cell_id])
                original_length = len(new_stale_links)
                for refresher_cell_id in stale_links[stale_cell_id]:
                    if refresher_cell_id not in stale_cells:
                        continue
                    new_stale_links |= stale_links[refresher_cell_id]
                new_stale_links.discard(stale_cell_id)
                stale_link_changes = stale_link_changes or original_length != len(
                    new_stale_links
                )
                stale_links[stale_cell_id] = new_stale_links
        for stale_cell_id in stale_cells:
            stale_links[stale_cell_id] -= stale_cells
            for refresher_cell_id in stale_links[stale_cell_id]:
                refresher_links[refresher_cell_id].add(stale_cell_id)
        return FrontendCheckerResult(
            # TODO: we should probably have separate fields for stale vs non-typechecking cells,
            #  or at least change the name to a more general "unsafe_cells" or equivalent
            stale_cells=stale_cells | typecheck_error_cells | unsafe_order_cells,
            fresh_cells=fresh_cells,
            new_fresh_cells=new_fresh_cells,
            forced_reactive_cells=forced_reactive_cells,
            stale_links=stale_links,
            refresher_links=refresher_links,
            phantom_cell_info=phantom_cell_info,
        )

    @staticmethod
    def _get_max_timestamp_cell_num_for_symbols(
        deep_symbols: Set[DataSymbol], shallow_symbols: Set[DataSymbol]
    ) -> int:
        max_timestamp_cell_num = -1
        for dsym in deep_symbols:
            max_timestamp_cell_num = max(
                max_timestamp_cell_num, dsym.timestamp.cell_num
            )
        for dsym in shallow_symbols:
            max_timestamp_cell_num = max(
                max_timestamp_cell_num, dsym.timestamp_excluding_ns_descendents.cell_num
            )
        return max_timestamp_cell_num

    def _safety_precheck_cell(self, cell: ExecutedCodeCell) -> None:
        checker_result = self.check_and_link_multiple_cells(
            cells_to_check=[cell],
            update_liveness_time_versions=self.mut_settings.static_slicing_enabled,
        )
        if cell.cell_id in checker_result.stale_cells:
            self.safety_issue_detected = True

    def _resync_symbols(self, symbols: Iterable[DataSymbol]):
        for dsym in symbols:
            if not dsym.containing_scope.is_global:
                continue
            obj = get_ipython().user_global_ns.get(dsym.name, None)
            if obj is None:
                continue
            if dsym.obj_id == id(obj):
                continue
            for alias in self.aliases[dsym.cached_obj_id] | self.aliases[dsym.obj_id]:
                containing_namespace = alias.containing_namespace
                if containing_namespace is None:
                    continue
                containing_obj = containing_namespace.obj
                if containing_obj is None:
                    continue
                # TODO: handle dict case too
                if isinstance(containing_obj, list) and containing_obj[-1] is obj:
                    containing_namespace._subscript_data_symbol_by_name.pop(
                        alias.name, None
                    )
                    alias.name = len(containing_obj) - 1
                    alias.update_obj_ref(obj)
                    containing_namespace._subscript_data_symbol_by_name[
                        alias.name
                    ] = alias
            self.aliases[dsym.cached_obj_id].discard(dsym)
            self.aliases[dsym.obj_id].discard(dsym)
            self.aliases[id(obj)].add(dsym)
            dsym.update_obj_ref(obj)

    def create_dag_metadata(
        self,
    ) -> Dict[int, Dict[str, Union[List[int], List[str], Dict[str, Dict[str, str]]]]]:
        cell_num_to_used_imports: Dict[int, Set[DataSymbol]] = defaultdict(set)
        cell_num_to_dynamic_inputs: Dict[int, Set[DataSymbol]] = defaultdict(set)
        cell_num_to_dynamic_outputs: Dict[int, Set[DataSymbol]] = defaultdict(set)
        cell_num_to_dynamic_cell_parents: Dict[int, Set[int]] = defaultdict(set)
        cell_num_to_dynamic_cell_children: Dict[int, Set[int]] = defaultdict(set)

        for sym in self.all_data_symbols():
            top_level_sym = sym.get_top_level()
            if (
                top_level_sym is None
                or not top_level_sym.is_globally_accessible
                or top_level_sym.is_anonymous
            ):
                # TODO: also skip lambdas
                continue
            for (
                used_time,
                sym_timestamp_when_used,
            ) in sym.timestamp_by_used_time.items():
                if top_level_sym.is_import:
                    cell_num_to_used_imports[used_time.cell_num].add(top_level_sym)
                elif used_time.cell_num != sym_timestamp_when_used.cell_num:
                    cell_num_to_dynamic_cell_parents[used_time.cell_num].add(
                        sym_timestamp_when_used.cell_num
                    )
                    cell_num_to_dynamic_cell_children[
                        sym_timestamp_when_used.cell_num
                    ].add(used_time.cell_num)
                    cell_num_to_dynamic_inputs[used_time.cell_num].add(top_level_sym)
                    cell_num_to_dynamic_outputs[sym_timestamp_when_used.cell_num].add(
                        top_level_sym
                    )
            if not top_level_sym.is_import:
                for updated_time in sym.updated_timestamps:
                    # TODO: distinguished between used / unused outputs?
                    cell_num_to_dynamic_outputs[updated_time.cell_num].add(
                        top_level_sym
                    )

        cell_metadata: Dict[
            int, Dict[str, Union[List[int], List[str], Dict[str, Dict[str, str]]]]
        ] = {}
        all_relevant_cells = (
            cell_num_to_used_imports.keys()
            | cell_num_to_dynamic_inputs.keys()
            | cell_num_to_dynamic_outputs.keys()
            | cell_num_to_dynamic_cell_parents.keys()
            | cell_num_to_dynamic_cell_children.keys()
        )
        for cell_num in all_relevant_cells:
            cell_imports = [
                dsym.get_import_string() for dsym in cell_num_to_used_imports[cell_num]
            ]
            input_symbols = {
                str(dsym): {"type": dsym.get_type_annotation_string()}
                for dsym in cell_num_to_dynamic_inputs[cell_num]
            }
            output_symbols = {
                str(dsym): {"type": dsym.get_type_annotation_string()}
                for dsym in cell_num_to_dynamic_outputs[cell_num]
            }
            parent_cells = list(cell_num_to_dynamic_cell_parents[cell_num])
            child_cells = list(cell_num_to_dynamic_cell_children[cell_num])
            cell_metadata[cell_num] = {
                "cell_imports": cell_imports,
                "input_symbols": input_symbols,
                "output_symbols": output_symbols,
                "parent_cells": parent_cells,
                "child_cells": child_cells,
            }
        return cell_metadata

    @contextmanager
    def _patch_pyccolo_exec(self):
        """
        The purpose of this context manager is to disable this project's
        tracer inside pyccolo's "exec()" functions, since it probably
        will not work properly inside of these.
        """
        orig_exec = pyc.exec
        orig_tracer_exec = pyc.BaseTracer.exec

        def _patched_exec(*args, **kwargs):
            with SafetyTracer.instance().tracing_disabled():
                return orig_exec(
                    *args,
                    num_extra_lookback_frames=kwargs.pop("num_extra_lookback_frames", 0)
                    + 1,
                    **kwargs,
                )

        def _patched_tracer_exec(*args, **kwargs):
            with SafetyTracer.instance().tracing_disabled():
                return orig_tracer_exec(*args, **kwargs)

        try:
            pyc.exec = _patched_exec
            pyc.BaseTracer.exec = orig_tracer_exec
            yield
        finally:
            pyc.exec = orig_exec
            pyc.BaseTracer.exec = orig_tracer_exec

    async def safe_execute(self, cell_content: str, is_async: bool, run_cell_func):
        if self._saved_debug_message is not None:  # pragma: no cover
            logger.error(self._saved_debug_message)
            self._saved_debug_message = None
        ret = None
        with save_number_of_currently_executing_cell():
            cell_id, self._active_cell_id = self._active_cell_id, None
            assert cell_id is not None
            cell = cells().create_and_track(
                cell_id,
                cell_content,
                self._tags,
                validate_ipython_counter=self.settings.store_history,
            )

            # Stage 1: Precheck.
            self._safety_precheck_cell(cell)

            # Stage 2: Trace / run the cell, updating dependencies as they are encountered.
            try:
                with self._tracing_context():
                    if is_async:
                        ret = await run_cell_func(cell_content)  # pragma: no cover
                    else:
                        ret = run_cell_func(cell_content)
                # Stage 2.1: resync any defined symbols that could have gotten out-of-sync
                #  due to tracing being disabled

                self._resync_symbols(
                    [
                        # TODO: avoid bad performance by only iterating over symbols updated in this cell
                        sym
                        for sym in self.all_data_symbols()
                        if sym.timestamp.cell_num == cells().exec_counter()
                    ]
                )
                self._gc()
            except Exception as e:
                if self.is_test:
                    self.set_exception_raised_during_execution(e)
            finally:
                return ret

    def _make_cell_magic(self, cell_magic_name):
        # this is to avoid capturing `self` and creating an extra reference to the singleton
        store_history = self.settings.store_history

        def _run_cell_func(cell):
            run_cell(cell, store_history=store_history)

        def _dependency_safety(_, cell: str):
            asyncio.get_event_loop().run_until_complete(
                asyncio.wait(
                    [singletons.nbs().safe_execute(cell, False, _run_cell_func)]
                )
            )

        # FIXME (smacke): probably not a great idea to rely on this
        _dependency_safety.__name__ = cell_magic_name
        return register_cell_magic(_dependency_safety)

    @contextmanager
    def _tracing_context(self):
        self.updated_symbols.clear()
        self.updated_reactive_symbols.clear()
        self.updated_deep_reactive_symbols.clear()

        with SafetyTracer.instance().tracing_context():
            SafetyTracer.instance().reset()
            ast_rewriter = SafetyTracer.instance().make_ast_rewriter(
                module_id=self.cell_counter()
            )
            with input_transformer_context(
                SafetyTracer.instance().make_syntax_augmenters(ast_rewriter)
                if self.settings.enable_reactive_modifiers
                else []
            ):
                with ast_transformer_context([ast_rewriter]):
                    with self._patch_pyccolo_exec():
                        yield

    def _make_line_magic(self):
        print_ = print  # to keep the test from failing since this is a legitimate print
        line_magic_names = [
            f[0] for f in inspect.getmembers(line_magics) if inspect.isfunction(f[1])
        ]

        def _handle(cmd, line):
            if cmd in ("deps", "show_deps", "show_dependency", "show_dependencies"):
                return line_magics.show_deps(line)
            elif cmd in ("stale", "show_stale"):
                return line_magics.show_stale(line)
            elif cmd == "trace_messages":
                return line_magics.trace_messages(line)
            elif cmd in ("hls", "nohls", "highlight", "highlights"):
                return line_magics.set_highlights(cmd, line)
            elif cmd in ("dag", "make_dag", "cell_dag", "make_cell_dag"):
                return json.dumps(self.create_dag_metadata(), indent=2)
            elif cmd in ("slice", "make_slice", "gather_slice"):
                return line_magics.make_slice(line)
            elif cmd in ("mode", "exec_mode"):
                return line_magics.set_exec_mode(line)
            elif cmd in ("schedule", "exec_schedule", "execution_schedule"):
                return line_magics.set_exec_schedule(line)
            elif cmd in ("flow", "flow_order", "semantics", "flow_semantics"):
                return line_magics.set_flow_order(line)
            elif cmd == "clear":
                self.min_timestamp = self.cell_counter()
                return None
            elif cmd in line_magic_names:
                logger.warning(
                    "We have a magic for %s, but have not yet registered it", cmd
                )
                return None
            else:
                logger.warning(line_magics.USAGE)
                return None

        def _safety(line: str):
            # this is to avoid capturing `self` and creating an extra reference to the singleton
            try:
                cmd, line = line.split(" ", 1)
                if cmd in ("slice", "make_slice", "gather_slice"):
                    # FIXME: hack to workaround some input transformer
                    line = re.sub(r"--tag +<class '(\w+)'>", r"--tag $\1", line)
            except ValueError:
                cmd, line = line, ""
            try:
                line, fname = line.split(">", 1)
            except ValueError:
                line, fname = line, None
            line = line.strip()
            if fname is not None:
                fname = fname.strip()

            outstr = _handle(cmd, line)
            if outstr is None:
                return

            if fname is None:
                print_(outstr)
            else:
                with open(fname, "w") as f:
                    f.write(outstr)

        # FIXME (smacke): probably not a great idea to rely on this
        _safety.__name__ = _SAFETY_LINE_MAGIC
        return register_line_magic(_safety)

    @property
    def dependency_tracking_enabled(self):
        return self.settings.track_dependencies

    @property
    def cell_magic_name(self):
        return self._cell_magic.__name__

    @property
    def line_magic_name(self):
        return self._line_magic.__name__

    def all_data_symbols(self) -> Iterable[DataSymbol]:
        for alias_set in self.aliases.values():
            yield from alias_set

    def test_and_clear_detected_flag(self):
        ret = self.safety_issue_detected
        self.safety_issue_detected = False
        return ret

    def _gc(self):
        # Need to do the garbage check and the collection separately
        garbage_syms = [
            dsym for dsym in self.all_data_symbols() if dsym.is_new_garbage()
        ]
        for dsym in garbage_syms:
            dsym.collect_self_garbage()

    def retrieve_namespace_attr_or_sub(
        self, obj: Any, attr_or_sub: SupportedIndexType, is_subscript: bool
    ):
        try:
            with pyc.allow_reentrant_event_handling():
                if is_subscript:
                    # TODO: more complete list of things that are checkable
                    #  or could cause side effects upon subscripting
                    return obj[attr_or_sub]
                else:
                    if self.is_develop:
                        assert isinstance(attr_or_sub, str)
                    return getattr(obj, cast(str, attr_or_sub))
        except (AttributeError, IndexError, KeyError):
            raise
        except Exception as e:
            if self.is_develop:
                logger.warning("unexpected exception: %s", e)
                logger.warning("object: %s", obj)
                logger.warning("attr / subscript: %s", attr_or_sub)
            raise e



# -*- coding: utf-8 -*-
import re
from typing import List, Set
from nbsafety.tracing.nbsafety_tracer import reactive_spec
from pyccolo.syntax_augmentation import (
    AugmentationType,
    AugmentationSpec,
    AUGMENTED_SYNTAX_REGEX_TEMPLATE,
    replace_tokens_and_get_augmented_positions,
)
from nbsafety.singletons import tracer
from .utils import make_safety_fixture

_safety_fixture, run_cell = make_safety_fixture(enable_reactive_modifiers=True)


REACTIVE_ATOM_REGEX = re.compile(
    AUGMENTED_SYNTAX_REGEX_TEMPLATE.format(token=reactive_spec.escaped_token)
)


def _replace_reactive_atoms(s: str) -> str:
    return replace_tokens_and_get_augmented_positions(
        s, reactive_spec, REACTIVE_ATOM_REGEX
    )[0]


def _get_reactive_positions(s: str) -> List[int]:
    return replace_tokens_and_get_augmented_positions(
        s, reactive_spec, REACTIVE_ATOM_REGEX
    )[1]


def _get_augmented_positions(s: str, spec: AugmentationSpec) -> List[int]:
    regex = re.compile(AUGMENTED_SYNTAX_REGEX_TEMPLATE.format(token=spec.escaped_token))
    return replace_tokens_and_get_augmented_positions(s, spec, regex)[1]


def _get_all_reactive_var_names() -> Set[str]:
    return {
        tracer().ast_node_by_id[node_id].id for node_id in tracer().reactive_node_ids
    }


def test_simple():
    assert REACTIVE_ATOM_REGEX.match("") is None
    assert REACTIVE_ATOM_REGEX.match("foo") is None
    assert REACTIVE_ATOM_REGEX.match("foo bar") is None
    assert _replace_reactive_atoms("$foo $bar") == "foo bar"
    assert _replace_reactive_atoms("$foo bar $baz42") == "foo bar baz42"
    assert _replace_reactive_atoms("$foo $42bar $_baz42") == "foo 42bar _baz42"


def test_simple_names_recovered():
    run_cell("x = 0")
    run_cell("y = $x + 1")
    assert _get_all_reactive_var_names() == {"x"}
    run_cell("z = $y + 2")
    assert _get_all_reactive_var_names() == {"x", "y"}
    run_cell("w1 = $z + 2\nw2 = $w1 + 3")
    assert _get_all_reactive_var_names() == {"x", "y", "z", "w1"}


def test_nested_names_recovered():
    run_cell(
        """
        def assert_nonzero(v):
            assert v != 0
        """
    )
    run_cell("x = 42")
    run_cell("$assert_nonzero($x)")
    varnames = _get_all_reactive_var_names()
    assert varnames == {"x", "assert_nonzero"}, "got %s" % varnames


def test_reactive_positions():
    assert _get_reactive_positions("foo") == []
    assert _get_reactive_positions("$foo") == [0]
    assert _get_reactive_positions("foo $bar") == [4]
    assert _get_reactive_positions("$foo $bar") == [0, 4]
    assert _get_reactive_positions("$foo $bar $baz") == [0, 4, 8]
    assert _get_reactive_positions("foo $bar $baz") == [4, 8]
    assert _get_reactive_positions("$foo bar $baz") == [0, 8]


def test_positions_with_offset_from_replacement():
    spec = AugmentationSpec(AugmentationType.prefix, "$$", "$")
    assert _get_augmented_positions("foo", spec) == []
    assert _get_augmented_positions("$$foo", spec) == [0]
    assert _get_augmented_positions("foo $$bar", spec) == [4]
    assert _get_augmented_positions("$$foo $$bar", spec) == [0, 5]
    assert _get_augmented_positions("$$foo $$bar $$baz", spec) == [0, 5, 10]
    assert _get_augmented_positions("foo $$bar $$baz", spec) == [4, 9]
    assert _get_augmented_positions("$$foo bar $$baz", spec) == [0, 9]



# -*- coding: utf-8 -*-
import ast
import sys
import textwrap
from typing import Set, Tuple, Union

from nbsafety.analysis.symbol_ref import SymbolRef
from nbsafety.analysis.live_refs import (
    compute_live_dead_symbol_refs as compute_live_dead_symbol_refs_with_stmts,
)
from .utils import make_safety_fixture


_safety_fixture, _ = make_safety_fixture()


def _simplify_symbol_refs(symbols: Set[SymbolRef]) -> Set[str]:
    simplified = set()
    for sym in symbols:
        if isinstance(sym, SymbolRef):
            sym = sym.chain[0].value
        if isinstance(sym, str):
            simplified.add(sym)
    return simplified


def compute_live_dead_symbol_refs(
    code: Union[str, ast.AST]
) -> Tuple[Set[str], Set[str]]:
    if isinstance(code, str):
        code = textwrap.dedent(code)
    live, dead = compute_live_dead_symbol_refs_with_stmts(code)
    live = {ref.ref for ref in live}
    live, dead = _simplify_symbol_refs(live), _simplify_symbol_refs(dead)
    return live, dead


def test_simple():
    live, dead = compute_live_dead_symbol_refs(
        """
        x = 5
        print(foo, x)
        """
    )
    assert live == {"foo", "print"}
    assert dead == {"x"}


def test_function_body():
    fbody = (
        ast.parse(
            textwrap.dedent(
                """
        def func():
            y = 42
            print(foo, bar, baz, x)
            x = 5
        """
            )
        )
        .body[0]
        .body
    )
    live, dead = compute_live_dead_symbol_refs(fbody)
    assert live == {"foo", "bar", "baz", "x", "print"}
    assert dead == {"x", "y"}


def test_comprehension_with_killed_elt():
    live, dead = compute_live_dead_symbol_refs(
        "[x for y in range(10) for x in range(11)]"
    )
    assert live == {"range"}
    assert dead == {"x", "y"}


def test_comprehension_with_live_elt():
    live, dead = compute_live_dead_symbol_refs(
        "[x for y in range(10) for _ in range(11)]"
    )
    assert live == {"x", "range"}, "got %s" % live
    assert dead == {"y", "_"}


def test_subscript_is_live():
    live, dead = compute_live_dead_symbol_refs("foo[bar] = baz")
    assert live == {"bar", "baz"}


if sys.version_info >= (3, 8):

    def test_walrus():
        live, dead = compute_live_dead_symbol_refs(
            """
            if (y := (x := x + 1) + 1) > 0:
                z = y + 1
            """
        )
        assert live == {"x"}
        assert dead == {"y", "z"}, "got %s" % dead
